import pandas as pd
from scipy.stats import chi2_contingency
import seaborn as sns
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
from anndata import AnnData




def roe(
        adata: AnnData,
        sample_key: str,
        cell_type_key: str,
        pval_threshold: float = 0.05,
        expected_value_threshold: float = 5,
        order="F"
) -> pd.DataFrame:
    """
    Calculate the ratio of observed cell number to expected cell number (Ro/e) for each cell type in each sample.

    Arguments:
        adata: AnnData object.
        sample_key: Key for sample information in adata.obs.
        cell_type_key: Key for cell type information in adata.obs.
        pval_threshold: Threshold for p-value. Default: 0.05.
        expected_value_threshold: Threshold for expected value. Default: 5.
        order: Order of columns in the contingency table. Default: "F".

    Returns:
        Ro/e: results in a DataFrame.
    
    """
    
    # Create a contingency table
    num_cell = pd.crosstab(index=adata.obs[cell_type_key], columns=adata.obs[sample_key])
    num_cell.index.name = "cluster"
    # If necessary, reorder columns based on the given order
    if order != "F":
        col_order = order.split(',')
        num_cell = num_cell[col_order]
    # Perform chi-square test
    chi2, p, dof, expected = chi2_contingency(num_cell)
    print(f"chi2: {chi2}, dof: {dof}, pvalue: {p}")
    # Check p-value
    if p <= pval_threshold:
        if all(item > expected_value_threshold for item in expected.flatten()):
            expected_data = pd.DataFrame(expected, index=num_cell.index, columns=num_cell.columns)
            roe = num_cell / expected_data
            adata.uns['roe_results'] = roe
            adata.uns['expected_values'] = expected_data
            adata.uns['sig_roe_results'] = roe
        else:
            print(
                f"Some expected frequencies are less than {expected_value_threshold}, it is suggested to use other statistical methods, such as Fisher's exact test")
            expected_data = pd.DataFrame(expected, index=num_cell.index, columns=num_cell.columns)
            adata.uns['expected_values'] = expected_data
            roe = num_cell / expected_data
            adata.uns['roe_results'] = roe
            adata.uns['unsig_roe_results'] = roe
    else:
        print("P-value is greater than 0.05, there is no statistical significance")
        roe_ratio = num_cell / expected
        adata.uns['unsig_roe_results'] = roe_ratio

    if 'sig_roe_results' in adata.uns:
        return adata.uns['sig_roe_results']
    else:
        return adata.uns['unsig_roe_results']



def roe_plot_heatmap(adata: AnnData, display_numbers: bool = False, center_value: float = 1.0,
                     color_scheme: str = 'cool',
                 custom_colors: list = None, save_path: str = None, batch_order: list = None):
    # 检查是否有显著的结果
    if 'sig_roe_results' in adata.uns:
        roe = adata.uns['sig_roe_results']
    else:
        roe = adata.uns['unsig_roe_results']

    # 如果提供了批次顺序，则按批次排序
    if batch_order:
        roe = roe.loc[batch_order]

    # 定义默认颜色方案
    color_schemes = {
        'default': ['#D73027', '#FFFFFF', '#1E682A'],
        'cool': ['#440154', '#FFFFFF', '#fde725'],
        'warm': ['#D53E4F', '#F46D43', '#FDAE61']
    }

    # 创建子图以便后续添加图例
    fig, ax = plt.subplots(figsize=(10, 7))

    # 创建定制的颜色映射，使1成为中间值
    if custom_colors:
        colors = custom_colors
    else:
        colors = color_schemes[color_scheme]

    custom_cmap = LinearSegmentedColormap.from_list('custom_cmap', colors, N=256)
    custom_cmap.set_bad(color='white')

    # 根据display_numbers参数选择显示方式
    if display_numbers:
        annot = roe.round(2)  # 保留两位小数
        sns.heatmap(roe, annot=annot, cmap=custom_cmap, center=center_value, fmt='', ax=ax)
        plt.title("ROE Results")
    else:
        transformed_roe = transform_roe_values(roe)
        sns.heatmap(roe, annot=transformed_roe, cmap=custom_cmap, center=center_value, fmt='', cbar=False, ax=ax)

        # 添加图例
        cbar_ax = fig.add_axes([0.93, 0.2, 0.03, 0.6])
        cbar = plt.colorbar(plt.cm.ScalarMappable(cmap=custom_cmap), cax=cbar_ax, orientation='vertical')
        cbar.set_ticks([0.125, 0.375, 0.625, 0.875])
        cbar.set_ticklabels(['+++', '++', '+', '±'])

    plt.title("Ro/e")
    if save_path:
        plt.savefig(save_path)  # 保存图像
    else:
        plt.show()


def transform_roe_values(roe):
    # 将roe DataFrame转换为字符串格式，用于标注
    transformed_roe = roe.copy()
    transformed_roe = transformed_roe.applymap(
        lambda x: '+++' if x >= 2 else ('++' if x >= 1.5 else ('+' if x >= 1 else '+/-')))
    return transformed_roe

# roe(adata, sample_key='batch', cell_type_key='celltypist_cell_label_coarse')
# plot_heatmap(adata, display_numbers=True)
# batch_order = ['sample3', 'sample1', 'sample2']
# plot_heatmap(adata, batch_order=batch_order)





from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Literal, overload
from warnings import warn

import anndata as ad
import numpy as np
from anndata import AnnData
from packaging.version import Version
from sklearn.utils import check_random_state

from .._settings import EMOJI



from scanpy import logging as logg
from scanpy._compat import DaskArray, pkg_version
from scanpy._settings import settings
from scanpy._utils import _doc_params, _empty, get_literal_vals, is_backed_type
from scanpy.get import _check_mask, _get_obs_rep
from scanpy.preprocessing._docs import doc_mask_var_hvg
from scanpy.preprocessing._pca._compat import _pca_compat_sparse
from scipy import sparse

if TYPE_CHECKING:
    from collections.abc import Container
    from collections.abc import Set as AbstractSet
    from typing import LiteralString, TypeVar

    import dask_ml.decomposition as dmld
    import sklearn.decomposition as skld
    from numpy.typing import DTypeLike, NDArray

    from scanpy._utils import Empty
    from scanpy._utils.random import _LegacyRandom

    MethodDaskML = type[dmld.PCA | dmld.IncrementalPCA | dmld.TruncatedSVD]
    MethodSklearn = type[skld.PCA | skld.TruncatedSVD]

    T = TypeVar("T", bound=LiteralString)
    M = TypeVar("M", bound=LiteralString)


SvdSolvPCADaskML = Literal["auto", "full", "tsqr", "randomized"]
SvdSolvTruncatedSVDDaskML = Literal["tsqr", "randomized"]
SvdSolvDaskML = SvdSolvPCADaskML | SvdSolvTruncatedSVDDaskML

if pkg_version("scikit-learn") >= Version("1.5") or TYPE_CHECKING:
    SvdSolvPCASparseSklearn = Literal["arpack", "covariance_eigh"]
else:
    SvdSolvPCASparseSklearn = Literal["arpack"]
SvdSolvPCADenseSklearn = Literal["auto", "full", "randomized"] | SvdSolvPCASparseSklearn
SvdSolvTruncatedSVDSklearn = Literal["arpack", "randomized"]
SvdSolvSkearn = (
    SvdSolvPCADenseSklearn | SvdSolvPCASparseSklearn | SvdSolvTruncatedSVDSklearn
)

SvdSolvPCACustom = Literal["covariance_eigh"]

SvdSolver = SvdSolvDaskML | SvdSolvSkearn | SvdSolvPCACustom

SpBase = sparse.spmatrix | sparse.sparray  # noqa: TID251
"""Only use when you directly convert it to a known subclass."""

_CSArray = sparse.csr_array | sparse.csc_array  # noqa: TID251
"""Only use if you want to specially handle arrays as opposed to matrices."""

_CSMatrix = sparse.csr_matrix | sparse.csc_matrix  # noqa: TID251
"""Only use if you want to specially handle matrices as opposed to arrays."""

CSRBase = sparse.csr_matrix | sparse.csr_array  # noqa: TID251
CSCBase = sparse.csc_matrix | sparse.csc_array  # noqa: TID251
CSBase = _CSArray | _CSMatrix


@_doc_params(
    mask_var_hvg=doc_mask_var_hvg,
)
def pca(  # noqa: PLR0912, PLR0913, PLR0915
    data: AnnData | np.ndarray | CSBase,
    n_comps: int | None = None,
    *,
    layer: str | None = None,
    zero_center: bool | None = True,
    svd_solver: SvdSolver | None = None,
    random_state: _LegacyRandom = 0,
    return_info: bool = False,
    mask_var: NDArray[np.bool_] | str | None | Empty = _empty,
    use_highly_variable: bool | None = None,
    dtype: DTypeLike = "float32",
    chunked: bool = False,
    chunk_size: int | None = None,
    key_added: str | None = None,
    copy: bool = False,
    use_gpu: bool = False,
) -> AnnData | np.ndarray | CSBase | None:
    r"""Principal component analysis :cite:p:`Pedregosa2011`.

    Computes PCA coordinates, loadings and variance decomposition.
    Uses the implementation of *scikit-learn* :cite:p:`Pedregosa2011`.

    .. versionchanged:: 1.5.0

        In previous versions, computing a PCA on a sparse matrix would make
        a dense copy of the array for mean centering.
        As of scanpy 1.5.0, mean centering is implicit.
        While results are extremely similar, they are not exactly the same.
        If you would like to reproduce the old results, pass a dense array.

    Parameters
    ----------
    data
        The (annotated) data matrix of shape `n_obs` × `n_vars`.
        Rows correspond to cells and columns to genes.
    n_comps
        Number of principal components to compute. Defaults to 50, or 1 - minimum
        dimension size of selected representation.
    layer
        If provided, which element of layers to use for PCA.
    zero_center
        If `True`, compute standard PCA from covariance matrix.
        If `False`, omit zero-centering variables
        (uses *scikit-learn* :class:`~sklearn.decomposition.TruncatedSVD` or
        *dask-ml* :class:`~dask_ml.decomposition.TruncatedSVD`),
        which allows to handle sparse input efficiently.
        Passing `None` decides automatically based on sparseness of the data.
    svd_solver
        SVD solver to use:

        `None`
            See `chunked` and `zero_center` descriptions to determine which class will be used.
            Depending on the class and the type of X different values for default will be set.
            For sparse *dask* arrays, will use `'covariance_eigh'`.
            If *scikit-learn* :class:`~sklearn.decomposition.PCA` is used, will give `'arpack'`,
            if *scikit-learn* :class:`~sklearn.decomposition.TruncatedSVD` is used, will give `'randomized'`,
            if *dask-ml* :class:`~dask_ml.decomposition.PCA` or :class:`~dask_ml.decomposition.IncrementalPCA` is used, will give `'auto'`,
            if *dask-ml* :class:`~dask_ml.decomposition.TruncatedSVD` is used, will give `'tsqr'`
        `'arpack'`
            for the ARPACK wrapper in SciPy (:func:`~scipy.sparse.linalg.svds`)
            Not available with *dask* arrays.
        `'covariance_eigh'`
            Classic eigendecomposition of the covariance matrix, suited for tall-and-skinny matrices.
            With dask, array must be CSR or dense and chunked as (N, adata.shape[1]).
        `'randomized'`
            for the randomized algorithm due to Halko (2009). For *dask* arrays,
            this will use :func:`~dask.array.linalg.svd_compressed`.
        `'auto'`
            chooses automatically depending on the size of the problem.
        `'tsqr'`
            Only available with dense *dask* arrays. "tsqr"
            algorithm from Benson et. al. (2013).

        .. versionchanged:: 1.9.3
           Default value changed from `'arpack'` to None.
        .. versionchanged:: 1.4.5
           Default value changed from `'auto'` to `'arpack'`.

        Efficient computation of the principal components of a sparse matrix
        currently only works with the `'arpack`' or `'covariance_eigh`' solver.

        If X is a sparse *dask* array, a custom `'covariance_eigh'` solver will be used.
        If X is a dense *dask* array, *dask-ml* classes :class:`~dask_ml.decomposition.PCA`,
        :class:`~dask_ml.decomposition.IncrementalPCA`, or
        :class:`~dask_ml.decomposition.TruncatedSVD` will be used.
        Otherwise their *scikit-learn* counterparts :class:`~sklearn.decomposition.PCA`,
        :class:`~sklearn.decomposition.IncrementalPCA`, or
        :class:`~sklearn.decomposition.TruncatedSVD` will be used.
    random_state
        Change to use different initial states for the optimization.
    return_info
        Only relevant when not passing an :class:`~anndata.AnnData`:
        see “Returns”.
    {mask_var_hvg}
    layer
        Layer of `adata` to use as expression values.
    dtype
        Numpy data type string to which to convert the result.
    chunked
        If `True`, perform an incremental PCA on segments of `chunk_size`.
        The incremental PCA automatically zero centers and ignores settings of
        `random_seed` and `svd_solver`. Uses sklearn :class:`~sklearn.decomposition.IncrementalPCA` or
        *dask-ml* :class:`~dask_ml.decomposition.IncrementalPCA`. If `False`, perform a full PCA and
        use sklearn :class:`~sklearn.decomposition.PCA` or
        *dask-ml* :class:`~dask_ml.decomposition.PCA`
    chunk_size
        Number of observations to include in each chunk.
        Required if `chunked=True` was passed.
    key_added
        If not specified, the embedding is stored as
        :attr:`~anndata.AnnData.obsm`\ `['X_pca']`, the loadings as
        :attr:`~anndata.AnnData.varm`\ `['PCs']`, and the the parameters in
        :attr:`~anndata.AnnData.uns`\ `['pca']`.
        If specified, the embedding is stored as
        :attr:`~anndata.AnnData.obsm`\ ``[key_added]``, the loadings as
        :attr:`~anndata.AnnData.varm`\ ``[key_added]``, and the the parameters in
        :attr:`~anndata.AnnData.uns`\ ``[key_added]``.
    copy
        If an :class:`~anndata.AnnData` is passed, determines whether a copy
        is returned. Is ignored otherwise.

    Returns
    -------
    If `data` is array-like and `return_info=False` was passed,
    this function returns the PCA representation of `data` as an
    array of the same type as the input array.

    Otherwise, it returns `None` if `copy=False`, else an updated `AnnData` object.
    Sets the following fields:

    `.obsm['X_pca' | key_added]` : :class:`~scipy.sparse.csr_matrix` | :class:`~scipy.sparse.csc_matrix` | :class:`~numpy.ndarray` (shape `(adata.n_obs, n_comps)`)
        PCA representation of data.
    `.varm['PCs' | key_added]` : :class:`~numpy.ndarray` (shape `(adata.n_vars, n_comps)`)
        The principal components containing the loadings.
    `.uns['pca' | key_added]['variance_ratio']` : :class:`~numpy.ndarray` (shape `(n_comps,)`)
        Ratio of explained variance.
    `.uns['pca' | key_added]['variance']` : :class:`~numpy.ndarray` (shape `(n_comps,)`)
        Explained variance, equivalent to the eigenvalues of the
        covariance matrix.

    """
    logg_start = logg.info(f"computing PCA{EMOJI['start']}")
    if layer is not None and chunked:
        # Current chunking implementation relies on pca being called on X
        msg = "Cannot use `layer` and `chunked` at the same time."
        raise NotImplementedError(msg)

    # chunked calculation is not randomized, anyways
    if svd_solver in {"auto", "randomized"} and not chunked:
        logg.info(
            f"{EMOJI['warning']} scikit-learn's randomized PCA might not be exactly "
            "reproducible across different computational platforms. For exact "
            "reproducibility, choose `svd_solver='arpack'`."
        )
    if return_anndata := isinstance(data, AnnData):
        if layer is None and not chunked and is_backed_type(data.X):
            msg = f"PCA is not implemented for matrices of type {type(data.X)} with chunked as False"
            raise NotImplementedError(msg)
        adata = data.copy() if copy else data
    elif pkg_version("anndata") < Version("0.8.0rc1"):
        adata = AnnData(data, dtype=data.dtype)
    else:
        adata = AnnData(data)

    # Unify new mask argument and deprecated use_highly_varible argument
    mask_var_param, mask_var = _handle_mask_var(adata, mask_var, use_highly_variable)
    del use_highly_variable
    adata_comp = adata[:, mask_var] if mask_var is not None else adata

    if n_comps is None:
        min_dim = min(adata_comp.n_vars, adata_comp.n_obs)
        n_comps = min_dim - 1 if min_dim <= settings.N_PCS else settings.N_PCS

    logg.info(f"    with {n_comps=}")

    X = _get_obs_rep(adata_comp, layer=layer)
    if is_backed_type(X) and layer is not None:
        msg = f"PCA is not implemented for matrices of type {type(X)} from layers"
        raise NotImplementedError(msg)
    # See: https://github.com/scverse/scanpy/pull/2816#issuecomment-1932650529
    if (
        Version(ad.__version__) < Version("0.9")
        and mask_var is not None
        and isinstance(X, np.ndarray)
    ):
        warnings.warn(
            "When using a mask parameter with anndata<0.9 on a dense array, the PCA"
            "can have slightly different results due the array being column major "
            "instead of row major.",
            UserWarning,
            stacklevel=2,
        )

    # check_random_state returns a numpy RandomState when passed an int but
    # dask needs an int for random state
    if not isinstance(X, DaskArray):
        random_state = random_state
    elif not isinstance(random_state, int):
        msg = f"random_state needs to be an int, not a {type(random_state).__name__} when passing a dask array"
        raise TypeError(msg)

    if chunked:
        if (
            not zero_center
            or random_state
            or (svd_solver is not None and svd_solver != "arpack")
        ):
            logg.debug("Ignoring zero_center, random_state, svd_solver")

        incremental_pca_kwargs = dict()
        if use_gpu and not isinstance(X, DaskArray):
            from torchdr import IncrementalPCA
            from numpy import zeros
            import torch
            import gc
            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            X_pca = zeros((X.shape[0], n_comps), X.dtype)
            torch.cuda.reset_peak_memory_stats(device)
            pca_ = IncrementalPCA(n_components=n_comps, device=device, 
                                  batch_size=chunk_size,
                                  **incremental_pca_kwargs)
            pca_.fit(X, check_input=True)
            X_pca = pca_.transform(X)
            # Reset GPU memory stats before fitting.
            
            del pca_
            torch.cuda.empty_cache()
            gc.collect()

        else:
            if isinstance(X, DaskArray):
                from dask.array import zeros
                from dask_ml.decomposition import IncrementalPCA

                incremental_pca_kwargs["svd_solver"] = _handle_dask_ml_args(
                    svd_solver, IncrementalPCA
                )
            else:
                from numpy import zeros
                from sklearn.decomposition import IncrementalPCA

            X_pca = zeros((X.shape[0], n_comps), X.dtype)
            pca_ = IncrementalPCA(n_components=n_comps, **incremental_pca_kwargs)

            for chunk, _, _ in adata_comp.chunked_X(chunk_size):
                chunk_dense = chunk.toarray() if isinstance(chunk, CSBase) else chunk
                pca_.partial_fit(chunk_dense)

            for chunk, start, end in adata_comp.chunked_X(chunk_size):
                chunk_dense = chunk.toarray() if isinstance(chunk, CSBase) else chunk
                X_pca[start:end] = pca_.transform(chunk_dense)
        
    elif zero_center:
        if isinstance(X, CSBase) and (
            pkg_version("scikit-learn") < Version("1.4") or svd_solver == "lobpcg"
        ):
            if svd_solver not in (
                {"lobpcg"} | get_literal_vals(SvdSolvPCASparseSklearn)
            ):
                if svd_solver is not None:
                    msg = (
                        f"Ignoring {svd_solver=} and using 'arpack', "
                        "sparse PCA with sklearn < 1.4 only supports 'lobpcg' and 'arpack'."
                    )
                    warnings.warn(msg, UserWarning, stacklevel=2)
                svd_solver = "arpack"
            elif svd_solver == "lobpcg":
                msg = (
                    f"{svd_solver=} for sparse relies on legacy code and will not be supported in the future. "
                    "Also the lobpcg solver has been observed to be inaccurate. Please use 'arpack' instead."
                )
                warnings.warn(msg, FutureWarning, stacklevel=2)
            X_pca, pca_ = _pca_compat_sparse(
                X, n_comps, solver=svd_solver, random_state=random_state
            )
        else:
            if not isinstance(X, DaskArray):
                if use_gpu:
                    if svd_solver == "auto":
                        svd_solver = "gesvd"
                    if svd_solver not in ["gesvd", "gesvdj", "gesvda"]:
                        svd_solver = "gesvd"
                    from torchdr import  PCA
                    import torch
                    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
                    pca_ = PCA(
                        n_components=n_comps,
                        device=device,
                        svd_driver=svd_solver,
                        random_state=random_state,
                    )
                from sklearn.decomposition import PCA

                svd_solver = _handle_sklearn_args(
                    svd_solver, PCA, sparse=isinstance(X, CSBase)
                )
                pca_ = PCA(
                    n_components=n_comps,
                    svd_solver=svd_solver,
                    random_state=random_state,
                )
            elif isinstance(X._meta, CSBase) or svd_solver == "covariance_eigh":
                from ._dask import PCAEighDask

                if random_state != 0:
                    msg = f"Ignoring {random_state=} when using a sparse dask array"
                    warnings.warn(msg, UserWarning, stacklevel=2)
                if svd_solver not in {None, "covariance_eigh"}:
                    msg = f"Ignoring {svd_solver=} when using a sparse dask array"
                    warnings.warn(msg, UserWarning, stacklevel=2)
                pca_ = PCAEighDask(n_components=n_comps)
            else:
                from dask_ml.decomposition import PCA

                svd_solver = _handle_dask_ml_args(svd_solver, PCA)
                pca_ = PCA(
                    n_components=n_comps,
                    svd_solver=svd_solver,
                    random_state=random_state,
                )
            X_pca = pca_.fit_transform(X)
    else:
        if isinstance(X, DaskArray):
            if isinstance(X._meta, CSBase):
                msg = "Dask sparse arrays do not support zero-centering (yet)"
                raise TypeError(msg)
            from dask_ml.decomposition import TruncatedSVD

            svd_solver = _handle_dask_ml_args(svd_solver, TruncatedSVD)
        else:
            from sklearn.decomposition import TruncatedSVD

            svd_solver = _handle_sklearn_args(svd_solver, TruncatedSVD)

        logg.debug(
            "    without zero-centering: \n"
            "    the explained variance does not correspond to the exact statistical definition\n"
            "    the first component, e.g., might be heavily influenced by different means\n"
            "    the following components often resemble the exact PCA very closely"
        )
        pca_ = TruncatedSVD(
            n_components=n_comps, random_state=random_state, algorithm=svd_solver
        )
        X_pca = pca_.fit_transform(X)

    if X_pca.dtype.descr != np.dtype(dtype).descr:
        X_pca = X_pca.astype(dtype)

    if return_anndata:
        key_obsm, key_varm, key_uns = (
            ("X_pca", "PCs", "pca") if key_added is None else [key_added] * 3
        )
        adata.obsm[key_obsm] = X_pca

        if mask_var is not None:
            adata.varm[key_varm] = np.zeros(shape=(adata.n_vars, n_comps))
            adata.varm[key_varm][mask_var] = pca_.components_.T
        else:
            adata.varm[key_varm] = pca_.components_.T

        params = dict(
            zero_center=zero_center,
            use_highly_variable=mask_var_param == "highly_variable",
            mask_var=mask_var_param,
        )
        if layer is not None:
            params["layer"] = layer
        adata.uns[key_uns] = dict(
            params=params,
            variance=pca_.explained_variance_,
            variance_ratio=pca_.explained_variance_ratio_,
        )

        logg.info(f"    finished{EMOJI['done']}", time=logg_start)
        logg.debug(
            "and added\n"
            f"    {key_obsm!r}, the PCA coordinates (adata.obs)\n"
            f"    {key_varm!r}, the loadings (adata.varm)\n"
            f"    'pca_variance', the variance / eigenvalues (adata.uns[{key_uns!r}])\n"
            f"    'pca_variance_ratio', the variance ratio (adata.uns[{key_uns!r}])"
        )
        return adata if copy else None
    else:
        logg.info(f"    finished{EMOJI['done']}", time=logg_start)
        if return_info:
            return (
                X_pca,
                pca_.components_,
                pca_.explained_variance_ratio_,
                pca_.explained_variance_,
            )
        else:
            return X_pca


def _handle_mask_var(
    adata: AnnData,
    mask_var: NDArray[np.bool_] | str | Empty | None,
    use_highly_variable: bool | None,
) -> tuple[np.ndarray | str | None, np.ndarray | None]:
    """Unify new mask argument and deprecated use_highly_varible argument.

    Returns both the normalized mask parameter and the validated mask array.
    """
    # First, verify and possibly warn
    if use_highly_variable is not None:
        hint = (
            'Use_highly_variable=True can be called through mask_var="highly_variable". '
            "Use_highly_variable=False can be called through mask_var=None"
        )
        msg = f"Argument `use_highly_variable` is deprecated, consider using the mask argument. {hint}"
        warn(msg, FutureWarning, stacklevel=2)
        if mask_var is not _empty:
            msg = f"These arguments are incompatible. {hint}"
            raise ValueError(msg)

    # Handle default case and explicit use_highly_variable=True
    if use_highly_variable or (
        use_highly_variable is None
        and mask_var is _empty
        and "highly_variable" in adata.var.columns
    ):
        mask_var = "highly_variable"

    # Without highly variable genes, we don’t use a mask by default
    if mask_var is _empty or mask_var is None:
        return None, None
    return mask_var, _check_mask(adata, mask_var, "var")


@overload
def _handle_dask_ml_args(
    svd_solver: str | None, method: type[dmld.PCA | dmld.IncrementalPCA]
) -> SvdSolvPCADaskML: ...
@overload
def _handle_dask_ml_args(
    svd_solver: str | None, method: type[dmld.TruncatedSVD]
) -> SvdSolvTruncatedSVDDaskML: ...
def _handle_dask_ml_args(svd_solver: str | None, method: MethodDaskML) -> SvdSolvDaskML:
    import dask_ml.decomposition as dmld

    args: AbstractSet[SvdSolvDaskML]
    default: SvdSolvDaskML
    match method:
        case dmld.PCA | dmld.IncrementalPCA:
            args = get_literal_vals(SvdSolvPCADaskML)
            default = "auto"
        case dmld.TruncatedSVD:
            args = get_literal_vals(SvdSolvTruncatedSVDDaskML)
            default = "tsqr"
        case _:
            msg = f"Unknown {method=} in _handle_dask_ml_args"
            raise ValueError(msg)
    return _handle_x_args(svd_solver, method, args, default)


@overload
def _handle_sklearn_args(
    svd_solver: str | None, method: type[skld.TruncatedSVD], *, sparse: None = None
) -> SvdSolvTruncatedSVDSklearn: ...
@overload
def _handle_sklearn_args(
    svd_solver: str | None, method: type[skld.PCA], *, sparse: Literal[False]
) -> SvdSolvPCADenseSklearn: ...
@overload
def _handle_sklearn_args(
    svd_solver: str | None, method: type[skld.PCA], *, sparse: Literal[True]
) -> SvdSolvPCASparseSklearn: ...
def _handle_sklearn_args(
    svd_solver: str | None, method: MethodSklearn, *, sparse: bool | None = None
) -> SvdSolvSkearn:
    import sklearn.decomposition as skld

    args: AbstractSet[SvdSolvSkearn]
    default: SvdSolvSkearn
    suffix = ""
    match (method, sparse):
        case (skld.TruncatedSVD, None):
            args = get_literal_vals(SvdSolvTruncatedSVDSklearn)
            default = "randomized"
        case (skld.PCA, False):
            args = get_literal_vals(SvdSolvPCADenseSklearn)
            default = "arpack"
        case (skld.PCA, True):
            args = get_literal_vals(SvdSolvPCASparseSklearn)
            default = "arpack"
            suffix = " (with sparse input)"
        case _:
            msg = f"Unknown {method=} ({sparse=}) in _handle_sklearn_args"
            raise ValueError(msg)

    return _handle_x_args(svd_solver, method, args, default, suffix=suffix)


def _handle_x_args(
    svd_solver: str | None,
    method: type,
    args: Container[T],
    default: T,
    *,
    suffix: str = "",
) -> T:
    if svd_solver in args:
        return svd_solver
    if svd_solver is not None:
        msg = (
            f"Ignoring {svd_solver=} and using {default}, "
            f"{method.__module__}.{method.__qualname__}{suffix} only supports {args}."
        )
        # (4: caller of `pca` -> 3: `pca` -> 2: `_handle_{sklearn,dask_ml}_args` -> 1: here)
        warnings.warn(msg, UserWarning, stacklevel=4)
    return default


from __future__ import annotations

import random
import re
from typing import TYPE_CHECKING, Literal

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scanpy as sc
import seaborn as sns
from anndata import AnnData
from lamin_utils import logger
from mudata import MuData
import patsy
from inmoose import edgepy

from pertpy._doc import _doc_params, doc_common_plot_args

if TYPE_CHECKING:
    from collections.abc import Sequence
    from matplotlib.axes import Axes
    from matplotlib.colors import Colormap
    from matplotlib.figure import Figure

from scipy.sparse import csr_matrix
from sklearn.metrics.pairwise import euclidean_distances


class Milo:
    """Python implementation of Milo."""

    def __init__(self):
        # No need to import R packages anymore
        pass

    def load(
        self,
        input: AnnData,
        feature_key: str | None = "rna",
    ) -> MuData:
        """Prepare a MuData object for subsequent processing.

        Args:
            input: AnnData
            feature_key: Key to store the cell-level AnnData object in the MuData object

        Returns:
            :class:`mudata.MuData` object with original AnnData.

        Examples:
            >>> import pertpy as pt
            >>> adata = pt.dt.bhattacherjee()
            >>> milo = pt.tl.Milo()
            >>> mdata = milo.load(adata)

        """
        mdata = MuData({feature_key: input, "milo": AnnData()})
        return mdata

    def make_nhoods(
        self,
        data: AnnData | MuData,
        neighbors_key: str | None = None,
        feature_key: str | None = "rna",
        prop: float = 0.1,
        seed: int = 0,
        copy: bool = False,
    ):
        """Randomly sample vertices on a KNN graph to define neighbourhoods of cells.

        The set of neighborhoods get refined by computing the median profile for the neighbourhood in reduced dimensional space
        and by selecting the nearest vertex to this position.
        Thus, multiple neighbourhoods may be collapsed to prevent over-sampling the graph space.

        Args:
            data: AnnData object with KNN graph defined in `obsp` or MuData object with a modality with KNN graph defined in `obsp`
            neighbors_key: The key in `adata.obsp` or `mdata[feature_key].obsp` to use as KNN graph.
                           If not specified, `make_nhoods` looks .obsp['connectivities'] for connectivities (default storage places for `scanpy.pp.neighbors`).
                           If specified, it looks at .obsp[.uns[neighbors_key]['connectivities_key']] for connectivities.
            feature_key: If input data is MuData, specify key to cell-level AnnData object.
            prop: Fraction of cells to sample for neighbourhood index search.
            seed: Random seed for cell sampling.
            copy: Determines whether a copy of the `adata` is returned.

        Returns:
            If `copy=True`, returns the copy of `adata` with the result in `.obs`, `.obsm`, and `.uns`.
            Otherwise:

            nhoods: scipy.sparse._csr.csr_matrix in `adata.obsm['nhoods']`.
            A binary matrix of cell to neighbourhood assignments. Neighbourhoods in the columns are ordered by the order of the index cell in adata.obs_names

            nhood_ixs_refined: pandas.Series in `adata.obs['nhood_ixs_refined']`.
            A boolean indicating whether a cell is an index for a neighbourhood

            nhood_kth_distance: pandas.Series in `adata.obs['nhood_kth_distance']`.
            The distance to the kth nearest neighbour for each index cell (used for SpatialFDR correction)

            nhood_neighbors_key: `adata.uns["nhood_neighbors_key"]`
            KNN graph key, used for neighbourhood construction

        Examples:
            >>> import pertpy as pt
            >>> import scanpy as sc
            >>> adata = pt.dt.bhattacherjee()
            >>> milo = pt.tl.Milo()
            >>> mdata = milo.load(adata)
            >>> sc.pp.neighbors(mdata["rna"])
            >>> milo.make_nhoods(mdata["rna"])

        """
        if isinstance(data, MuData):
            adata = data[feature_key]
        if isinstance(data, AnnData):
            adata = data
        if copy:
            adata = adata.copy()

        # Get reduced dim used for KNN graph
        if neighbors_key is None:
            try:
                use_rep = adata.uns["neighbors"]["params"]["use_rep"]
            except KeyError:
                logger.warning("Using X_pca as default embedding")
                use_rep = "X_pca"
            try:
                knn_graph = adata.obsp["connectivities"].copy()
            except KeyError:
                logger.error('No "connectivities" slot in adata.obsp -- please run scanpy.pp.neighbors(adata) first')
                raise
        else:
            try:
                use_rep = adata.uns[neighbors_key]["params"]["use_rep"]
            except KeyError:
                logger.warning("Using X_pca as default embedding")
                use_rep = "X_pca"
            knn_graph = adata.obsp[neighbors_key + "_connectivities"].copy()

        X_dimred = adata.obsm[use_rep]
        n_ixs = int(np.round(adata.n_obs * prop))
        knn_graph[knn_graph != 0] = 1
        random.seed(seed)
        random_vertices = random.sample(range(adata.n_obs), k=n_ixs)
        random_vertices.sort()
        ixs_nn = knn_graph[random_vertices, :]
        non_zero_rows = ixs_nn.nonzero()[0]
        non_zero_cols = ixs_nn.nonzero()[1]
        refined_vertices = np.empty(
            shape=[
                len(random_vertices),
            ]
        )

        for i in range(len(random_vertices)):
            nh_pos = np.median(X_dimred[non_zero_cols[non_zero_rows == i], :], 0).reshape(-1, 1)
            nn_ixs = non_zero_cols[non_zero_rows == i]
            # Find closest real point (amongst nearest neighbors)
            dists = euclidean_distances(X_dimred[non_zero_cols[non_zero_rows == i], :], nh_pos.T)
            # Update vertex index
            refined_vertices[i] = nn_ixs[dists.argmin()]

        refined_vertices = np.unique(refined_vertices.astype("int"))
        refined_vertices.sort()

        nhoods = knn_graph[:, refined_vertices]
        adata.obsm["nhoods"] = nhoods

        # Add ixs to adata
        adata.obs["nhood_ixs_random"] = adata.obs_names.isin(adata.obs_names[random_vertices])
        adata.obs["nhood_ixs_refined"] = adata.obs_names.isin(adata.obs_names[refined_vertices])
        adata.obs["nhood_ixs_refined"] = adata.obs["nhood_ixs_refined"].astype("int")
        adata.obs["nhood_ixs_random"] = adata.obs["nhood_ixs_random"].astype("int")
        adata.uns["nhood_neighbors_key"] = neighbors_key
        # Store distance to K-th nearest neighbor (used for spatial FDR correction)
        knn_dists = adata.obsp["distances"] if neighbors_key is None else adata.obsp[neighbors_key + "_distances"]

        nhood_ixs = adata.obs["nhood_ixs_refined"] == 1
        dist_mat = knn_dists[np.asarray(nhood_ixs), :]
        k_distances = dist_mat.max(1).toarray().ravel()
        adata.obs["nhood_kth_distance"] = 0
        adata.obs["nhood_kth_distance"] = adata.obs["nhood_kth_distance"].astype(float)
        adata.obs.loc[adata.obs["nhood_ixs_refined"] == 1, "nhood_kth_distance"] = k_distances

        if copy:
            return adata

    def count_nhoods(
        self,
        data: AnnData | MuData,
        sample_col: str,
        feature_key: str | None = "rna",
    ):
        """Builds a sample-level AnnData object storing the matrix of cell counts per sample per neighbourhood.

        Args:
            data: AnnData object with neighbourhoods defined in `obsm['nhoods']` or MuData object with a modality with neighbourhoods defined in `obsm['nhoods']`
            sample_col: Column in adata.obs that contains sample information
            feature_key: If input data is MuData, specify key to cell-level AnnData object.

        Returns:
            MuData object storing the original (i.e. rna) AnnData in `mudata[feature_key]`
            and the compositional anndata storing the neighbourhood cell counts in `mudata['milo']`.
            Here:
            - `mudata['milo'].obs_names` are samples (defined from `adata.obs['sample_col']`)
            - `mudata['milo'].var_names` are neighbourhoods
            - `mudata['milo'].X` is the matrix counting the number of cells from each
            sample in each neighbourhood

        Examples:
            >>> import pertpy as pt
            >>> import scanpy as sc
            >>> adata = pt.dt.bhattacherjee()
            >>> milo = pt.tl.Milo()
            >>> mdata = milo.load(adata)
            >>> sc.pp.neighbors(mdata["rna"])
            >>> milo.make_nhoods(mdata["rna"])
            >>> mdata = milo.count_nhoods(mdata, sample_col="orig.ident")

        """
        if isinstance(data, MuData):
            adata = data[feature_key]
            is_MuData = True
        if isinstance(data, AnnData):
            adata = data
            is_MuData = False
        if isinstance(adata, AnnData):
            try:
                nhoods = adata.obsm["nhoods"]
            except KeyError:
                logger.error('Cannot find "nhoods" slot in adata.obsm -- please run milopy.make_nhoods(adata)')
                raise
        # Make nhood abundance matrix
        sample_dummies = pd.get_dummies(adata.obs[sample_col])
        all_samples = sample_dummies.columns
        sample_dummies = csr_matrix(sample_dummies.values)
        nhood_count_mat = nhoods.T.dot(sample_dummies)
        sample_obs = pd.DataFrame(index=all_samples)
        sample_adata = AnnData(X=nhood_count_mat.T, obs=sample_obs)
        sample_adata.uns["sample_col"] = sample_col
        # Save nhood index info
        sample_adata.var["index_cell"] = adata.obs_names[adata.obs["nhood_ixs_refined"] == 1]
        sample_adata.var["kth_distance"] = adata.obs.loc[
            adata.obs["nhood_ixs_refined"] == 1, "nhood_kth_distance"
        ].values

        if is_MuData is True:
            data.mod["milo"] = sample_adata
            return data
        else:
            milo_mdata = MuData({feature_key: adata, "milo": sample_adata})
            return milo_mdata

    def da_nhoods(
        self,
        mdata: MuData,
        design: str,
        model_contrasts: str | None = None,
        subset_samples: list[str] | None = None,
        add_intercept: bool = True,
        feature_key: str | None = "rna",
        solver: Literal["edger", "batchglm"] = "edger",
    ):
        """Performs differential abundance testing on neighbourhoods using QLF test implementation.

        Args:
            mdata: MuData object
            design: Formula for the test, following glm syntax from R (e.g. '~ condition').
                    Terms should be columns in `milo_mdata[feature_key].obs`.
            model_contrasts: A string vector that defines the contrasts used to perform DA testing, following glm syntax from R (e.g. "conditionDisease - conditionControl").
                             If no contrast is specified (default), then the last categorical level in condition of interest is used as the test group.
            subset_samples: subset of samples (obs in `milo_mdata['milo']`) to use for the test.
            add_intercept: whether to include an intercept in the model. If False, this is equivalent to adding + 0 in the design formula. When model_contrasts is specified, this is set to False by default.
            feature_key: If input data is MuData, specify key to cell-level AnnData object.
            solver: The solver to fit the model to. One of "edger" or "batchglm" (currently only "edger" is implemented)

        Returns:
            None, modifies `milo_mdata['milo']` in place, adding the results of the DA test to `.var`:
            - `logFC` stores the log fold change in cell abundance (coefficient from the GLM)
            - `PValue` stores the p-value for the QLF test before multiple testing correction
            - `SpatialFDR` stores the p-value adjusted for multiple testing to limit the false discovery rate,
                calculated with weighted Benjamini-Hochberg procedure

        Examples:
            >>> import pertpy as pt
            >>> import scanpy as sc
            >>> adata = pt.dt.bhattacherjee()
            >>> milo = pt.tl.Milo()
            >>> mdata = milo.load(adata)
            >>> sc.pp.neighbors(mdata["rna"])
            >>> milo.make_nhoods(mdata["rna"])
            >>> mdata = milo.count_nhoods(mdata, sample_col="orig.ident")
            >>> milo.da_nhoods(mdata, design="~label")

        """
        try:
            sample_adata = mdata["milo"]
        except KeyError:
            logger.error(
                "milo_mdata should be a MuData object with two slots:"
                " feature_key and 'milo' - please run milopy.count_nhoods() first"
            )
            raise
        adata = mdata[feature_key]

        covariates = [x.strip(" ") for x in set(re.split("\\+|\\*", design.lstrip("~ ")))]

        # Add covariates used for testing to sample_adata.var
        sample_col = sample_adata.uns["sample_col"]
        try:
            sample_obs = adata.obs[covariates + [sample_col]].drop_duplicates()
        except KeyError:
            missing_cov = [x for x in covariates if x not in sample_adata.obs.columns]
            logger.warning("Covariates {c} are not columns in adata.obs".format(c=" ".join(missing_cov)))
            raise
        sample_obs = sample_obs[covariates + [sample_col]]
        sample_obs.index = sample_obs[sample_col].astype("str")

        try:
            assert sample_obs.loc[sample_adata.obs_names].shape[0] == len(sample_adata.obs_names)
        except AssertionError:
            logger.warning(
                f"Values in mdata[{feature_key}].obs[{covariates}] cannot be unambiguously assigned to each sample"
                f" -- each sample value should match a single covariate value"
            )
            raise
        sample_adata.obs = sample_obs.loc[sample_adata.obs_names]

        # Get design dataframe
        try:
            design_df = sample_adata.obs[covariates]
        except KeyError:
            missing_cov = [x for x in covariates if x not in sample_adata.obs.columns]
            logger.error(
                'Covariates {c} are not columns in adata.uns["sample_adata"].obs'.format(c=" ".join(missing_cov))
            )
            raise
        # Get count matrix
        count_mat = sample_adata.X.T.toarray()
        lib_size = count_mat.sum(0)

        # Filter out samples with zero counts
        keep_smp = lib_size > 0

        # Subset samples
        if subset_samples is not None:
            keep_smp = keep_smp & sample_adata.obs_names.isin(subset_samples)
            design_df = design_df[keep_smp]
            for i, e in enumerate(design_df.columns):
                if design_df.dtypes[i].name == "category":
                    design_df[e] = design_df[e].cat.remove_unused_categories()

        # Filter out nhoods with zero counts (they can appear after sample filtering)
        keep_nhoods = count_mat[:, keep_smp].sum(1) > 0

        if solver == "edger":
            # Define model matrix using patsy
            if not add_intercept or model_contrasts is not None:
                design_formula = design + " + 0"
            else:
                design_formula = design
            
            # Create model matrix
            model_matrix = patsy.dmatrix(design_formula, design_df, return_type='dataframe')
            
            # Create DGEList object
            dge = edgepy.DGEList(
                counts=count_mat[keep_nhoods, :][:, keep_smp],
                lib_size=lib_size[keep_smp]
            )
            
            # Normalize
            dge = edgepy.calcNormFactors(dge, method="TMM")
            
            # Estimate dispersion
            dge = edgepy.estimateDisp(dge, design=model_matrix)
            
            # Fit GLM QLF
            fit = edgepy.glmQLFit(dge, design=model_matrix, robust=True)
            
            # Test
            if model_contrasts is not None:
                # Parse contrasts
                # For edgepy, we need to create the contrast matrix manually
                contrast_cols = list(model_matrix.columns)
                
                # Create contrast vector
                contrast_vector = np.zeros(len(contrast_cols))
                
                # Parse the contrast string (e.g., "conditionDisease - conditionControl")
                contrast_terms = re.split(r'\s*[+-]\s*', model_contrasts.strip())
                signs = [1] + [-1 if s == '-' else 1 for s in re.findall(r'[+-]', model_contrasts)]
                
                for term, sign in zip(contrast_terms, signs):
                    term = term.strip()
                    if term in contrast_cols:
                        idx = contrast_cols.index(term)
                        contrast_vector[idx] = sign
                    else:
                        logger.error(f"Contrast term '{term}' not found in model matrix columns")
                        raise ValueError(f"Invalid contrast term: {term}")
                
                # Perform test with contrast
                result = edgepy.glmQLFTest(fit, contrast=contrast_vector)
                res = edgepy.topTags(result, n=np.inf, sort_by="none")
            else:
                # Test last coefficient
                n_coef = model_matrix.shape[1]
                result = edgepy.glmQLFTest(fit, coef=n_coef-1)  # Python uses 0-based indexing
                res = edgepy.topTags(result, n=np.inf, sort_by="none")
            
            # Convert to DataFrame if not already
            if not isinstance(res, pd.DataFrame):
                res = pd.DataFrame(res)
        
        elif solver == "batchglm":
            raise NotImplementedError("batchglm solver is not yet implemented")

        # Save outputs
        res.index = sample_adata.var_names[keep_nhoods]
        if any(col in sample_adata.var.columns for col in res.columns):
            sample_adata.var = sample_adata.var.drop(res.columns, axis=1)
        sample_adata.var = pd.concat([sample_adata.var, res], axis=1)

        # Run Graph spatial FDR correction
        self._graph_spatial_fdr(sample_adata, neighbors_key=adata.uns["nhood_neighbors_key"])

    def annotate_nhoods(
        self,
        mdata: MuData,
        anno_col: str,
        feature_key: str | None = "rna",
    ):
        """Assigns a categorical label to neighbourhoods, based on the most frequent label among cells in each neighbourhood. This can be useful to stratify DA testing results by cell types or samples.

        Args:
            mdata: MuData object
            anno_col: Column in adata.obs containing the cell annotations to use for nhood labelling
            feature_key: If input data is MuData, specify key to cell-level AnnData object.

        Returns:
            Adds in place.
            - `milo_mdata['milo'].var["nhood_annotation"]`: assigning a label to each nhood
            - `milo_mdata['milo'].var["nhood_annotation_frac"]` stores the fraciton of cells in the neighbourhood with the assigned label
            - `milo_mdata['milo'].varm['frac_annotation']`: stores the fraction of cells from each label in each nhood
            - `milo_mdata['milo'].uns["annotation_labels"]`: stores the column names for `milo_mdata['milo'].varm['frac_annotation']`

        Examples:
            >>> import pertpy as pt
            >>> import scanpy as sc
            >>> adata = pt.dt.bhattacherjee()
            >>> milo = pt.tl.Milo()
            >>> mdata = milo.load(adata)
            >>> sc.pp.neighbors(mdata["rna"])
            >>> milo.make_nhoods(mdata["rna"])
            >>> mdata = milo.count_nhoods(mdata, sample_col="orig.ident")
            >>> milo.annotate_nhoods(mdata, anno_col="cell_type")

        """
        try:
            sample_adata = mdata["milo"]
        except KeyError:
            logger.error(
                "milo_mdata should be a MuData object with two slots: feature_key and 'milo' - please run milopy.count_nhoods(adata) first"
            )
            raise
        adata = mdata[feature_key]

        # Check value is not numeric
        if pd.api.types.is_numeric_dtype(adata.obs[anno_col]):
            raise ValueError(
                "adata.obs[anno_col] is not of categorical type - please use milopy.utils.annotate_nhoods_continuous for continuous variables"
            )

        anno_dummies = pd.get_dummies(adata.obs[anno_col])
        anno_count = adata.obsm["nhoods"].T.dot(csr_matrix(anno_dummies.values))
        anno_count_dense = anno_count.toarray()
        anno_sum = anno_count_dense.sum(1)
        anno_frac = np.divide(anno_count_dense, anno_sum[:, np.newaxis])

        anno_frac_dataframe = pd.DataFrame(anno_frac, columns=anno_dummies.columns, index=sample_adata.var_names)
        sample_adata.varm["frac_annotation"] = anno_frac_dataframe.values
        sample_adata.uns["annotation_labels"] = anno_frac_dataframe.columns
        sample_adata.uns["annotation_obs"] = anno_col
        sample_adata.var["nhood_annotation"] = anno_frac_dataframe.idxmax(1)
        sample_adata.var["nhood_annotation_frac"] = anno_frac_dataframe.max(1)

    def annotate_nhoods_continuous(self, mdata: MuData, anno_col: str, feature_key: str | None = "rna"):
        """Assigns a continuous value to neighbourhoods, based on mean cell level covariate stored in adata.obs. This can be useful to correlate DA log-foldChanges with continuous covariates such as pseudotime, gene expression scores etc...

        Args:
            mdata: MuData object
            anno_col: Column in adata.obs containing the cell annotations to use for nhood labelling
            feature_key: If input data is MuData, specify key to cell-level AnnData object.

        Returns:
            Adds in place.
            - `milo_mdata['milo'].var["nhood_{anno_col}"]`: assigning a continuous value to each nhood

        Examples:
            >>> import pertpy as pt
            >>> import scanpy as sc
            >>> adata = pt.dt.bhattacherjee()
            >>> milo = pt.tl.Milo()
            >>> mdata = milo.load(adata)
            >>> sc.pp.neighbors(mdata["rna"])
            >>> milo.make_nhoods(mdata["rna"])
            >>> mdata = milo.count_nhoods(mdata, sample_col="orig.ident")
            >>> milo.annotate_nhoods_continuous(mdata, anno_col="nUMI")
        """
        if "milo" not in mdata.mod:
            raise ValueError(
                "milo_mdata should be a MuData object with two slots: feature_key and 'milo' - please run milopy.count_nhoods(adata) first"
            )
        adata = mdata[feature_key]

        # Check value is not categorical
        if not pd.api.types.is_numeric_dtype(adata.obs[anno_col]):
            raise ValueError(
                "adata.obs[anno_col] is not of continuous type - please use milopy.utils.annotate_nhoods for categorical variables"
            )

        anno_val = adata.obsm["nhoods"].T.dot(csr_matrix(adata.obs[anno_col]).T)

        mean_anno_val = anno_val.toarray() / np.array(adata.obsm["nhoods"].T.sum(1))

        mdata["milo"].var[f"nhood_{anno_col}"] = mean_anno_val

    def add_covariate_to_nhoods_var(self, mdata: MuData, new_covariates: list[str], feature_key: str | None = "rna"):
        """Add covariate from cell-level obs to sample-level obs. These should be covariates for which a single value can be assigned to each sample.

        Args:
            mdata: MuData object
            new_covariates: columns in `milo_mdata[feature_key].obs` to add to `milo_mdata['milo'].obs`.
            feature_key: If input data is MuData, specify key to cell-level AnnData object.

        Returns:
            None, adds columns to `milo_mdata['milo']` in place

        Examples:
            >>> import pertpy as pt
            >>> import scanpy as sc
            >>> adata = pt.dt.bhattacherjee()
            >>> milo = pt.tl.Milo()
            >>> mdata = milo.load(adata)
            >>> sc.pp.neighbors(mdata["rna"])
            >>> milo.make_nhoods(mdata["rna"])
            >>> mdata = milo.count_nhoods(mdata, sample_col="orig.ident")
            >>> milo.add_covariate_to_nhoods_var(mdata, new_covariates=["label"])
        """
        try:
            sample_adata = mdata["milo"]
        except KeyError:
            logger.error(
                "milo_mdata should be a MuData object with two slots: feature_key and 'milo' - please run milopy.count_nhoods(adata) first"
            )
            raise
        adata = mdata[feature_key]

        sample_col = sample_adata.uns["sample_col"]
        covariates = list(
            set(sample_adata.obs.columns[sample_adata.obs.columns != sample_col].tolist() + new_covariates)
        )
        try:
            sample_obs = adata.obs[covariates + [sample_col]].drop_duplicates()
        except KeyError:
            missing_cov = [covar for covar in covariates if covar not in sample_adata.obs.columns]
            logger.error("Covariates {c} are not columns in adata.obs".format(c=" ".join(missing_cov)))
            raise
        sample_obs = sample_obs[covariates + [sample_col]].astype("str")
        sample_obs.index = sample_obs[sample_col]
        try:
            assert sample_obs.loc[sample_adata.obs_names].shape[0] == len(sample_adata.obs_names)
        except ValueError:
            logger.error(
                "Covariates cannot be unambiguously assigned to each sample -- each sample value should match a single covariate value"
            )
            raise
        sample_adata.obs = sample_obs.loc[sample_adata.obs_names]

    def build_nhood_graph(self, mdata: MuData, basis: str = "X_umap", feature_key: str | None = "rna"):
        """Build graph of neighbourhoods used for visualization of DA results.

        Args:
            mdata: MuData object
            basis: Name of the obsm basis to use for layout of neighbourhoods (key in `adata.obsm`).
            feature_key: If input data is MuData, specify key to cell-level AnnData object.

        Returns:
            - `milo_mdata['milo'].varp['nhood_connectivities']`: graph of overlap between neighbourhoods (i.e. no of shared cells)
            - `milo_mdata['milo'].var["Nhood_size"]`: number of cells in neighbourhoods

        Examples:
            >>> import pertpy as pt
            >>> import scanpy as sc
            >>> adata = pt.dt.bhattacherjee()
            >>> milo = pt.tl.Milo()
            >>> mdata = milo.load(adata)
            >>> sc.pp.neighbors(mdata["rna"])
            >>> sc.tl.umap(mdata["rna"])
            >>> milo.make_nhoods(mdata["rna"])
            >>> mdata = milo.count_nhoods(mdata, sample_col="orig.ident")
            >>> milo.build_nhood_graph(mdata)
        """
        adata = mdata[feature_key]
        # # Add embedding positions
        mdata["milo"].varm["X_milo_graph"] = adata[adata.obs["nhood_ixs_refined"] == 1].obsm[basis]
        # Add nhood size
        mdata["milo"].var["Nhood_size"] = np.array(adata.obsm["nhoods"].sum(0)).flatten()
        # Add adjacency graph
        mdata["milo"].varp["nhood_connectivities"] = adata.obsm["nhoods"].T.dot(adata.obsm["nhoods"])
        mdata["milo"].varp["nhood_connectivities"].setdiag(0)
        mdata["milo"].varp["nhood_connectivities"].eliminate_zeros()
        mdata["milo"].uns["nhood"] = {
            "connectivities_key": "nhood_connectivities",
            "distances_key": "",
        }

    def add_nhood_expression(self, mdata: MuData, layer: str | None = None, feature_key: str | None = "rna") -> None:
        """Calculates the mean expression in neighbourhoods of each feature.

        Args:
            mdata: MuData object
            layer: If provided, use `milo_mdata[feature_key][layer]` as expression matrix instead of `milo_mdata[feature_key].X`.
            feature_key: If input data is MuData, specify key to cell-level AnnData object.

        Returns:
            Updates adata in place to store the matrix of average expression in each neighbourhood in `milo_mdata['milo'].varm['expr']`

        Examples:
            >>> import pertpy as pt
            >>> import scanpy as sc
            >>> adata = pt.dt.bhattacherjee()
            >>> milo = pt.tl.Milo()
            >>> mdata = milo.load(adata)
            >>> sc.pp.neighbors(mdata["rna"])
            >>> milo.make_nhoods(mdata["rna"])
            >>> mdata = milo.count_nhoods(mdata, sample_col="orig.ident")
            >>> milo.add_nhood_expression(mdata)

        """
        try:
            sample_adata = mdata["milo"]
        except KeyError:
            logger.error(
                "milo_mdata should be a MuData object with two slots:"
                " feature_key and 'milo' - please run milopy.count_nhoods(adata) first"
            )
            raise
        adata = mdata[feature_key]

        # Get gene expression matrix
        if layer is None:
            X = adata.X
            expr_id = "expr"
        else:
            X = adata.layers[layer]
            expr_id = "expr_" + layer

        # Aggregate over nhoods -- taking the mean
        nhoods_X = X.T.dot(adata.obsm["nhoods"])
        nhoods_X = csr_matrix(nhoods_X / adata.obsm["nhoods"].toarray().sum(0))
        sample_adata.varm[expr_id] = nhoods_X.T

    def _graph_spatial_fdr(
        self,
        sample_adata: AnnData,
        neighbors_key: str | None = None,
    ):
        """FDR correction weighted on inverse of connectivity of neighbourhoods.

        The distance to the k-th nearest neighbor is used as a measure of connectivity.

        Args:
            sample_adata: Sample-level AnnData.
            neighbors_key: The key in `adata.obsp` to use as KNN graph.
        """
        # use 1/connectivity as the weighting for the weighted BH adjustment from Cydar
        w = 1 / sample_adata.var["kth_distance"]
        w[np.isinf(w)] = 0

        # Computing a density-weighted q-value.
        pvalues = sample_adata.var["PValue"]
        keep_nhoods = ~pvalues.isna()  # Filtering in case of test on subset of nhoods
        o = pvalues[keep_nhoods].argsort()
        pvalues = pvalues.loc[keep_nhoods].iloc[o]
        w = w.loc[keep_nhoods].iloc[o]

        adjp = np.zeros(shape=len(o))
        adjp[o] = (sum(w) * pvalues / np.cumsum(w))[::-1].cummin()[::-1]
        adjp = np.array([x if x < 1 else 1 for x in adjp])

        sample_adata.var["SpatialFDR"] = np.nan
        sample_adata.var.loc[keep_nhoods, "SpatialFDR"] = adjp

    # Removed _setup_rpy2 and _try_import_bioc_library methods as they are no longer needed

    # The plotting methods remain unchanged
    @_doc_params(common_plot_args=doc_common_plot_args)
    def plot_nhood_graph(  # pragma: no cover # noqa: D417
        self,
        mdata: MuData,
        *,
        alpha: float = 0.1,
        min_logFC: float = 0,
        min_size: int = 10,
        plot_edges: bool = False,
        title: str = "DA log-Fold Change",
        color_map: Colormap | str | None = None,
        palette: str | Sequence[str] | None = None,
        ax: Axes | None = None,
        return_fig: bool = False,
        **kwargs,
    ) -> Figure | None:
        """Visualize DA results on abstracted graph (wrapper around sc.pl.embedding).

        Args:
            mdata: MuData object
            alpha: Significance threshold. (default: 0.1)
            min_logFC: Minimum absolute log-Fold Change to show results. If is 0, show all significant neighbourhoods.
            min_size: Minimum size of nodes in visualization. (default: 10)
            plot_edges: If edges for neighbourhood overlaps whould be plotted.
            title: Plot title.
            {common_plot_args}
            **kwargs: Additional arguments to `scanpy.pl.embedding`.

        Examples:
            >>> import pertpy as pt
            >>> import scanpy as sc
            >>> adata = pt.dt.bhattacherjee()
            >>> milo = pt.tl.Milo()
            >>> mdata = milo.load(adata)
            >>> sc.pp.neighbors(mdata["rna"])
            >>> sc.tl.umap(mdata["rna"])
            >>> milo.make_nhoods(mdata["rna"])
            >>> mdata = milo.count_nhoods(mdata, sample_col="orig.ident")
            >>> milo.da_nhoods(mdata,
            >>>            design='~label',
            >>>            model_contrasts='labelwithdraw_15d_Cocaine-labelwithdraw_48h_Cocaine')
            >>> milo.build_nhood_graph(mdata)
            >>> milo.plot_nhood_graph(mdata)

        Preview:
            .. image:: /_static/docstring_previews/milo_nhood_graph.png
        """
        nhood_adata = mdata["milo"].T.copy()

        if "Nhood_size" not in nhood_adata.obs.columns:
            raise KeyError(
                'Cannot find "Nhood_size" column in adata.uns["nhood_adata"].obs -- \
                    please run milopy.utils.build_nhood_graph(adata)'
            )

        nhood_adata.obs["graph_color"] = nhood_adata.obs["logFC"]
        nhood_adata.obs.loc[nhood_adata.obs["SpatialFDR"] > alpha, "graph_color"] = np.nan
        nhood_adata.obs["abs_logFC"] = abs(nhood_adata.obs["logFC"])
        nhood_adata.obs.loc[nhood_adata.obs["abs_logFC"] < min_logFC, "graph_color"] = np.nan

        # Plotting order - extreme logFC on top
        nhood_adata.obs.loc[nhood_adata.obs["graph_color"].isna(), "abs_logFC"] = np.nan
        ordered = nhood_adata.obs.sort_values("abs_logFC", na_position="first").index
        nhood_adata = nhood_adata[ordered]

        vmax = np.max([nhood_adata.obs["graph_color"].max(), abs(nhood_adata.obs["graph_color"].min())])
        vmin = -vmax

        fig = sc.pl.embedding(
            nhood_adata,
            "X_milo_graph",
            color="graph_color",
            cmap="RdBu_r",
            size=nhood_adata.obs["Nhood_size"] * min_size,
            edges=plot_edges,
            neighbors_key="nhood",
            sort_order=False,
            frameon=False,
            vmax=vmax,
            vmin=vmin,
            title=title,
            color_map=color_map,
            palette=palette,
            ax=ax,
            show=False,
            **kwargs,
        )

        if return_fig:
            return fig
        plt.show()
        return None

    @_doc_params(common_plot_args=doc_common_plot_args)
    def plot_nhood(  # pragma: no cover # noqa: D417
        self,
        mdata: MuData,
        ix: int,
        *,
        feature_key: str | None = "rna",
        basis: str = "X_umap",
        color_map: Colormap | str | None = None,
        palette: str | Sequence[str] | None = None,
        ax: Axes | None = None,
        return_fig: bool = False,
        **kwargs,
    ) -> Figure | None:
        """Visualize cells in a neighbourhood.

        Args:
            mdata: MuData object with feature_key slot, storing neighbourhood assignments in `mdata[feature_key].obsm['nhoods']`
            ix: index of neighbourhood to visualize
            feature_key: Key in mdata to the cell-level AnnData object.
            basis: Embedding to use for visualization.
            color_map: Colormap to use for coloring.
            palette: Color palette to use for coloring.
            ax: Axes to plot on.
            {common_plot_args}
            **kwargs: Additional arguments to `scanpy.pl.embedding`.

        Examples:
            >>> import pertpy as pt
            >>> import scanpy as sc
            >>> adata = pt.dt.bhattacherjee()
            >>> milo = pt.tl.Milo()
            >>> mdata = milo.load(adata)
            >>> sc.pp.neighbors(mdata["rna"])
            >>> sc.tl.umap(mdata["rna"])
            >>> milo.make_nhoods(mdata["rna"])
            >>> milo.plot_nhood(mdata, ix=0)

        Preview:
            .. image:: /_static/docstring_previews/milo_nhood.png
        """
        mdata[feature_key].obs["Nhood"] = mdata[feature_key].obsm["nhoods"][:, ix].toarray().ravel()
        fig = sc.pl.embedding(
            mdata[feature_key],
            basis,
            color="Nhood",
            size=30,
            title="Nhood" + str(ix),
            color_map=color_map,
            palette=palette,
            return_fig=return_fig,
            ax=ax,
            show=False,
            **kwargs,
        )

        if return_fig:
            return fig
        plt.show()
        return None

    @_doc_params(common_plot_args=doc_common_plot_args)
    def plot_da_beeswarm(  # pragma: no cover # noqa: D417
        self,
        mdata: MuData,
        *,
        feature_key: str | None = "rna",
        anno_col: str = "nhood_annotation",
        alpha: float = 0.1,
        subset_nhoods: list[str] = None,
        palette: str | Sequence[str] | dict[str, str] | None = None,
        return_fig: bool = False,
    ) -> Figure | None:
        """Plot beeswarm plot of logFC against nhood labels.

        Args:
            mdata: MuData object
            feature_key: Key in mdata to the cell-level AnnData object.
            anno_col: Column in adata.uns['nhood_adata'].obs to use as annotation. (default: 'nhood_annotation'.)
            alpha: Significance threshold. (default: 0.1)
            subset_nhoods: List of nhoods to plot. If None, plot all nhoods.
            palette: Name of Seaborn color palette for violinplots.
                     Defaults to pre-defined category colors for violinplots.
            {common_plot_args}

        Returns:
            If `return_fig` is `True`, returns the figure, otherwise `None`.

        Examples:
            >>> import pertpy as pt
            >>> import scanpy as sc
            >>> adata = pt.dt.bhattacherjee()
            >>> milo = pt.tl.Milo()
            >>> mdata = milo.load(adata)
            >>> sc.pp.neighbors(mdata["rna"])
            >>> milo.make_nhoods(mdata["rna"])
            >>> mdata = milo.count_nhoods(mdata, sample_col="orig.ident")
            >>> milo.da_nhoods(mdata, design="~label")
            >>> milo.annotate_nhoods(mdata, anno_col="cell_type")
            >>> milo.plot_da_beeswarm(mdata)

        Preview:
            .. image:: /_static/docstring_previews/milo_da_beeswarm.png
        """
        try:
            nhood_adata = mdata["milo"].T.copy()
        except KeyError:
            raise RuntimeError(
                "mdata should be a MuData object with two slots: feature_key and 'milo'. Run 'milopy.count_nhoods(adata)' first."
            ) from None

        try:
            nhood_adata.obs[anno_col]
        except KeyError:
            raise RuntimeError(
                f"Unable to find {anno_col} in mdata['milo'].var. Run 'milopy.utils.annotate_nhoods(adata, anno_col)' first"
            ) from None

        if subset_nhoods is not None:
            nhood_adata = nhood_adata[nhood_adata.obs[anno_col].isin(subset_nhoods)]

        try:
            nhood_adata.obs["logFC"]
        except KeyError:
            raise RuntimeError(
                "Unable to find 'logFC' in mdata.uns['nhood_adata'].obs. Run 'core.da_nhoods(adata)' first."
            ) from None

        sorted_annos = (
            nhood_adata.obs[[anno_col, "logFC"]].groupby(anno_col).median().sort_values("logFC", ascending=True).index
        )

        anno_df = nhood_adata.obs[[anno_col, "logFC", "SpatialFDR"]].copy()
        anno_df["is_signif"] = anno_df["SpatialFDR"] < alpha
        anno_df = anno_df[anno_df[anno_col] != "nan"]

        try:
            obs_col = nhood_adata.uns["annotation_obs"]
            if palette is None:
                palette = dict(
                    zip(
                        mdata[feature_key].obs[obs_col].cat.categories,
                        mdata[feature_key].uns[f"{obs_col}_colors"],
                        strict=False,
                    )
                )
            sns.violinplot(
                data=anno_df,
                y=anno_col,
                x="logFC",
                order=sorted_annos,
                inner=None,
                orient="h",
                palette=palette,
                linewidth=0,
                scale="width",
            )
        except BaseException:  # noqa: BLE001
            sns.violinplot(
                data=anno_df,
                y=anno_col,
                x="logFC",
                order=sorted_annos,
                inner=None,
                orient="h",
                linewidth=0,
                scale="width",
            )
        sns.stripplot(
            data=anno_df,
            y=anno_col,
            x="logFC",
            order=sorted_annos,
            size=2,
            hue="is_signif",
            palette=["grey", "black"],
            orient="h",
            alpha=0.5,
        )
        plt.legend(loc="upper left", title=f"< {int(alpha * 100)}% SpatialFDR", bbox_to_anchor=(1, 1), frameon=False)
        plt.axvline(x=0, ymin=0, ymax=1, color="black", linestyle="--")

        if return_fig:
            return plt.gcf()
        plt.show()
        return None

    @_doc_params(common_plot_args=doc_common_plot_args)
    def plot_nhood_counts_by_cond(  # pragma: no cover # noqa: D417
        self,
        mdata: MuData,
        test_var: str,
        *,
        subset_nhoods: list[str] = None,
        log_counts: bool = False,
        return_fig: bool = False,
    ) -> Figure | None:
        """Plot boxplot of cell numbers vs condition of interest.

        Args:
            mdata: MuData object storing cell level and nhood level information
            test_var: Name of column in adata.obs storing condition of interest (y-axis for boxplot)
            subset_nhoods: List of obs_names for neighbourhoods to include in plot. If None, plot all nhoods.
            log_counts: Whether to plot log1p of cell counts.
            {common_plot_args}

        Returns:
            If `return_fig` is `True`, returns the figure, otherwise `None`.
        """
        try:
            nhood_adata = mdata["milo"].T.copy()
        except KeyError:
            raise RuntimeError(
                "mdata should be a MuData object with two slots: feature_key and 'milo'. Run milopy.count_nhoods(mdata) first"
            ) from None

        if subset_nhoods is None:
            subset_nhoods = nhood_adata.obs_names

        pl_df = pd.DataFrame(nhood_adata[subset_nhoods].X.toarray(), columns=nhood_adata.var_names).melt(
            var_name=nhood_adata.uns["sample_col"], value_name="n_cells"
        )
        pl_df = pd.merge(pl_df, nhood_adata.var)
        pl_df["log_n_cells"] = np.log1p(pl_df["n_cells"])
        if not log_counts:
            sns.boxplot(data=pl_df, x=test_var, y="n_cells", color="lightblue")
            sns.stripplot(data=pl_df, x=test_var, y="n_cells", color="black", s=3)
            plt.ylabel("# cells")
        else:
            sns.boxplot(data=pl_df, x=test_var, y="log_n_cells", color="lightblue")
            sns.stripplot(data=pl_df, x=test_var, y="log_n_cells", color="black", s=3)
            plt.ylabel("log(# cells + 1)")

        plt.xticks(rotation=90)
        plt.xlabel(test_var)

        if return_fig:
            return plt.gcf()
        plt.show()
        return None


import numpy as np
import torch
import torch.nn.functional as F
from torch import nn
import os
import anndata as ad
import scanpy as sc
from typing import List


class Encoder(nn.Module):
    """A class that encapsulates the encoder."""
    def __init__(
        self,
        n_genes: int,
        latent_dim: int = 128,
        hidden_dim: List[int] = [1024, 1024],
        dropout: float = 0.5,
        input_dropout: float = 0.4,
        residual: bool = False,
    ):
        """Constructor.

        Parameters
        ----------
        n_genes: int
            The number of genes in the gene space, representing the input dimensions.
        latent_dim: int, default: 128
            The latent space dimensions
        hidden_dim: List[int], default: [1024, 1024]
            A list of hidden layer dimensions, describing the number of layers and their dimensions.
            Hidden layers are constructed in the order of the list for the encoder and in reverse
            for the decoder.
        dropout: float, default: 0.5
            The dropout rate for hidden layers
        input_dropout: float, default: 0.4
            The dropout rate for the input layer
        residual: bool, default: False
            Use residual connections.
        """
        super().__init__()
        self.latent_dim = latent_dim
        self.network = nn.ModuleList()
        self.residual = residual
        if self.residual:
            assert len(set(hidden_dim)) == 1
        for i in range(len(hidden_dim)):
            if i == 0:  # input layer
                self.network.append(
                    nn.Sequential(
                        nn.Dropout(p=input_dropout),
                        nn.Linear(n_genes, hidden_dim[i]),
                        nn.BatchNorm1d(hidden_dim[i]),
                        nn.PReLU(),
                    )
                )
            else:  # hidden layers
                self.network.append(
                    nn.Sequential(
                        nn.Dropout(p=dropout),
                        nn.Linear(hidden_dim[i - 1], hidden_dim[i]),
                        nn.BatchNorm1d(hidden_dim[i]),
                        nn.PReLU(),
                    )
                )
        # output layer
        self.network.append(nn.Linear(hidden_dim[-1], latent_dim))

    def forward(self, x) -> F.Tensor:
        for i, layer in enumerate(self.network):
            if self.residual and (0 < i < len(self.network) - 1):
                x = layer(x) + x
            else:
                x = layer(x)
        return F.normalize(x, p=2, dim=1)

    def save_state(self, filename: str):
        """Save state dictionary.

        Parameters
        ----------
        filename: str
            Filename to save the state dictionary.
        """
        torch.save({"state_dict": self.state_dict()}, filename)

    def load_state(self, filename: str, use_gpu: bool = False):
        """Load model state.

        Parameters
        ----------
        filename: str
            Filename containing the model state.
        use_gpu: bool
            Boolean indicating whether or not to use GPUs.
        """
        if not use_gpu:
            ckpt = torch.load(filename, map_location=torch.device("cpu"))
        else:
            ckpt = torch.load(filename)
        state_dict = ckpt['state_dict']
        first_layer_key = ['network.0.1.weight',
            'network.0.1.bias',
            'network.0.2.weight',
            'network.0.2.bias',
            'network.0.2.running_mean',
            'network.0.2.running_var',
            'network.0.2.num_batches_tracked',
            'network.0.3.weight]',]
        for key in first_layer_key:
            if key in state_dict:  
                del state_dict[key]
        self.load_state_dict(state_dict, strict=False)


class Decoder(nn.Module):
    """A class that encapsulates the decoder."""

    def __init__(
        self,
        n_genes: int,
        latent_dim: int = 128,
        hidden_dim: List[int] = [1024, 1024],
        dropout: float = 0.5,
        residual: bool = False,
    ):
        """Constructor.

        Parameters
        ----------
        n_genes: int
            The number of genes in the gene space, representing the input dimensions.
        latent_dim: int, default: 128
            The latent space dimensions
        hidden_dim: List[int], default: [1024, 1024]
            A list of hidden layer dimensions, describing the number of layers and their dimensions.
            Hidden layers are constructed in the order of the list for the encoder and in reverse
            for the decoder.
        dropout: float, default: 0.5
            The dropout rate for hidden layers
        residual: bool, default: False
            Use residual connections.
        """
        super().__init__()
        self.latent_dim = latent_dim
        self.network = nn.ModuleList()
        self.residual = residual
        if self.residual:
            assert len(set(hidden_dim)) == 1
        for i in range(len(hidden_dim)):
            if i == 0:  # first hidden layer
                self.network.append(
                    nn.Sequential(
                        nn.Linear(latent_dim, hidden_dim[i]),
                        nn.BatchNorm1d(hidden_dim[i]),
                        nn.PReLU(),
                    )
                )
            else:  # other hidden layers
                self.network.append(
                    nn.Sequential(
                        nn.Dropout(p=dropout),
                        nn.Linear(hidden_dim[i - 1], hidden_dim[i]),
                        nn.BatchNorm1d(hidden_dim[i]),
                        nn.PReLU(),
                    )
                )
        # reconstruction layer
        self.network.append(nn.Linear(hidden_dim[-1], n_genes))

    def forward(self, x):
        for i, layer in enumerate(self.network):
            if self.residual and (0 < i < len(self.network) - 1):
                x = layer(x) + x
            else:
                x = layer(x)
        return x

    def save_state(self, filename: str):
        """Save state dictionary.

        Parameters
        ----------
        filename: str
            Filename to save the state dictionary.
        """
        torch.save({"state_dict": self.state_dict()}, filename)

    def load_state(self, filename: str, use_gpu: bool = False):
        """Load model state.

        Parameters
        ----------
        filename: str
            Filename containing the model state.
        use_gpu: bool
            Boolean indicating whether to use GPUs.
        """
        if not use_gpu:
            ckpt = torch.load(filename, map_location=torch.device("cpu"))
        else:
            ckpt = torch.load(filename)
        state_dict = ckpt['state_dict']
        last_layer_key = ['network.3.weight',
                'network.3.bias',]
        for key in last_layer_key:
            if key in state_dict:  
                del state_dict[key]
        self.load_state_dict(state_dict, strict=False)
        # self.load_state_dict(ckpt["state_dict"])

class VAE(torch.nn.Module):
    """
    VAE base on compositional perturbation autoencoder (CPA)
    """
    def __init__(
        self,
        num_genes,
        device="cuda",
        seed=0,
        loss_ae="gauss",
        decoder_activation="linear",
        hidden_dim=128,
    ):
        super(VAE, self).__init__()
        # set generic attributes
        self.num_genes = num_genes
        self.device = device
        self.seed = seed
        self.loss_ae = loss_ae
        # early-stopping
        self.best_score = -1e3
        self.patience_trials = 0

        # set hyperparameters
        self.set_hparams_(hidden_dim)

        # set models
        self.hidden_dim = [1024,1024,1024]
        self.dropout = 0.0
        self.input_dropout = 0.0
        self.residual = False
        self.encoder = Encoder(
            self.num_genes,
            latent_dim=self.hparams["dim"],
            hidden_dim=self.hidden_dim,
            dropout=self.dropout,
            input_dropout=self.input_dropout,
            residual=self.residual,
        )
        self.decoder = Decoder(
            self.num_genes,
            latent_dim=self.hparams["dim"],
            hidden_dim=list(reversed(self.hidden_dim)),
            dropout=self.dropout,
            residual=self.residual,
        )

        # losses
        self.loss_autoencoder = nn.MSELoss(reduction='mean')

        self.iteration = 0

        self.to(self.device)

        # optimizers
        get_params = lambda model, cond: list(model.parameters()) if cond else []
        _parameters = (
            get_params(self.encoder, True)
            + get_params(self.decoder, True)
        )
        self.optimizer_autoencoder = torch.optim.AdamW(_parameters, lr=self.hparams["autoencoder_lr"], weight_decay=self.hparams["autoencoder_wd"],)


    def forward(self, genes, return_latent=False, return_decoded=False):
        """
        If return_latent=True, act as encoder only. If return_decoded, genes should 
        be the latent representation and this act as decoder only.
        """
        if return_decoded:
            gene_reconstructions = self.decoder(genes)
            gene_reconstructions = nn.ReLU()(gene_reconstructions)  # only relu when inference
            return gene_reconstructions

        latent_basal = self.encoder(genes)
        if return_latent:
            return latent_basal

        gene_reconstructions = self.decoder(latent_basal)

        return gene_reconstructions



    def set_hparams_(self, hidden_dim):
        """
        Set hyper-parameters to default values or values fixed by user.
        """

        self.hparams = {
            "dim": hidden_dim,
            "autoencoder_width": 5000,
            "autoencoder_depth": 3,
            "adversary_lr": 3e-4,
            "autoencoder_wd": 0.01, 
            "autoencoder_lr": 5e-4, 
        }

        return self.hparams


    def train(self, genes):
        """
        Train VAE.
        """
        genes = genes.to(self.device)
        gene_reconstructions = self.forward(genes)

        reconstruction_loss = self.loss_autoencoder(gene_reconstructions, genes)

        self.optimizer_autoencoder.zero_grad()
        reconstruction_loss.backward()
        self.optimizer_autoencoder.step()

        self.iteration += 1

        return {
            "loss_reconstruction": reconstruction_loss.item(),
        }



import argparse
from os import fspath
from pathlib import Path
import pandas as pd

import sys
from .cell_lineage_GRN import NetModel
from .cefcon_result_object import CefconResults
from .utils import data_preparation


def main():
    parser = argparse.ArgumentParser(prog='CEFCON', formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser = add_main_args(parser)
    args = parser.parse_args()

    ## output dir
    p = Path(args.out_dir)
    if not p.exists():
        Path.mkdir(p)

    ## load data
    data = data_preparation(args.input_expData, args.input_priorNet, genes_DE=args.input_genesDE,
                            additional_edges_pct=args.additional_edges_pct)
    data = data['all']

    ## GRN construction
    cefcon_GRN_model = NetModel(hidden_dim=args.hidden_dim,
                                output_dim=args.output_dim,
                                heads=args.heads,
                                attention_type=args.attention,
                                miu=args.miu,
                                epochs=args.epochs,
                                repeats=args.repeats,
                                seed=args.seed,
                                cuda=args.cuda,
                                )
    cefcon_GRN_model.run(data, showProgressBar=True)
    G_predicted = cefcon_GRN_model.get_network(keep_self_loops=~args.remove_self_loops,
                                               edge_threshold_avgDegree=args.edge_threshold_param,
                                               edge_threshold_zscore=None,
                                               output_file=fspath(p / 'cell_lineage_GRN.csv'))
    node_embeddings = cefcon_GRN_model.get_gene_embedding(output_file=fspath(p / 'gene_embs.csv'))
    cefcon_results = CefconResults(adata=cefcon_GRN_model._adata,
                                   network=G_predicted,
                                   gene_embedding=node_embeddings)

    ## Driver regulators
    cefcon_results.gene_influence_score()
    cefcon_results.driver_regulators(topK=args.topK_drivers, output_file=fspath(p / 'driver_regulators.csv'))

    ## RGMs
    RGMs_results_dict = cefcon_results.RGM_activity(return_value=True)
    RGMs = pd.DataFrame([{'Driver_Regulator': r.name, 'Members': list(r.gene2weight.keys())} for r in RGMs_results_dict['RGMs']])
    RGMs.to_csv(fspath(p / 'RGMs.csv'))
    RGMs_results_dict['aucell'].to_csv(fspath(p / 'AUCell_mtx.csv'))

    print('[Done!] Please check the results in "%s/"' % args.out_dir)


def add_main_args(parser: argparse.ArgumentParser):
    # Input data
    input_parser = parser.add_argument_group(title='Input data options')
    input_parser.add_argument('--input_expData', type=str, required=True, metavar='PATH',
                              help='path to the input gene expression data')
    input_parser.add_argument('--input_priorNet', type=str, required=True, metavar='PATH',
                              help='path to the input prior gene interaction network')
    input_parser.add_argument('--input_genesDE', type=str, default=None, metavar='PATH',
                              help='path to the input gene differential expression score')
    input_parser.add_argument('--additional_edges_pct', type=float, default=0.01,
                              help='proportion of high co-expression interactions to be added')

    # GRN
    grn_parser = parser.add_argument_group(title='Cell-lineage-specific GRN construction options')
    grn_parser.add_argument('--cuda', type=int, default=0,
                            help="an integer greater than -1 indicates the GPU device number and -1 indicates the CPU device")
    grn_parser.add_argument('--seed', type=int, default=2023,
                            help="random seed (set to -1 means no random seed is assigned)")

    grn_parser.add_argument("--hidden_dim", type=int, default=128,
                            help="hidden dimension of the GNN encoder")
    grn_parser.add_argument("--output_dim", type=int, default=64,
                            help="output dimension of the GNN encoder")
    grn_parser.add_argument("--heads", type=int, default=4,
                            help="number of heads")
    grn_parser.add_argument("--attention", type=str, default='COS', choices=['COS', 'AD', 'SD'],
                            help="type of attention scoring function")
    grn_parser.add_argument('--miu', type=float, default=0.5,
                            help='parameter for considering the importance of attention coefficients of the first GNN layer')
    grn_parser.add_argument('--epochs', type=int, default=350,
                            help='number of epochs for one run')
    grn_parser.add_argument('--repeats', type=int, default=5,
                            help='number of run repeats')

    grn_parser.add_argument("--edge_threshold_param", type=int, default=8,
                            help="threshold for selecting top-weighted edges (larger values means more edges)")
    grn_parser.add_argument("--remove_self_loops", action="store_true",
                            help="remove self loops")

    # Driver regulators
    driver_parser = parser.add_argument_group(title='Driver regulator identification options')
    driver_parser.add_argument('--topK_drivers', type=int, default=100,
                               help="number of top-ranked candidate driver genes according to their influence scores")

    # Output dir
    parser.add_argument("--out_dir", type=str, required=True, default='./output',
                        help="results output path")

    return parser


if __name__ == "__main__":
    main()



from __future__ import annotations

import logging
import os

import joblib
import numpy as np
import pandas as pd
from sklearn import svm
from sklearn.calibration import CalibratedClassifierCV

from popv import settings
from popv.algorithms._base_algorithm import BaseAlgorithm


class Support_Vector(BaseAlgorithm):
    """
    Compute a linear SVM classifier.

    Parameters
    ----------
    batch_key : str or None
        Key for batch information in adata.obs; default: "_batch_annotation".
    labels_key : str or None
        Key for cell-type information in adata.obs; default: "_labels_annotation".
    layer_key : str or None
        Key for the layer to use for classification. Defaults to None (using adata.X).
    result_key : str or None
        Key in adata.obs for storing predictions; default: "popv_svm_prediction".
    classifier_dict : dict or None
        Dictionary with non-default parameters for SVM. Defaults to:
        {'C': 1, 'max_iter': 5000, 'class_weight': 'balanced'}.
    train_both : bool
        Whether to train both cuml and sklearn classifiers.
    """

    def __init__(
        self,
        batch_key: str | None = "_batch_annotation",
        labels_key: str | None = "_labels_annotation",
        layer_key: str | None = None,
        result_key: str | None = "popv_svm_prediction",
        classifier_dict: dict | None = None,
        train_both: bool = False,
    ) -> None:
        super().__init__(batch_key=batch_key, labels_key=labels_key, result_key=result_key)
        self.layer_key = layer_key
        self.classifier_dict = {"C": 1, "max_iter": 5000, "class_weight": "balanced"}
        if classifier_dict:
            self.classifier_dict.update(classifier_dict)
        self.train_both = train_both

    def _save_model(self, clf, filename: str) -> None:
        save_path = os.path.join(adata.uns["_save_path_trained_models"], filename)
        with open(save_path, "wb") as f:
            joblib.dump(clf, f)

    def _load_model(self, filename: str):
        load_path = os.path.join(adata.uns["_save_path_trained_models"], filename)
        with open(load_path, "rb") as f:
            return joblib.load(f)

    def predict(self, adata):
        """
        Predict celltypes using a linear SVM.

        Parameters
        ----------
        adata
            Anndata object. Predictions are stored in adata.obs[self.result_key].
        """
        logging.info(f'Computing support vector machine. Storing prediction in adata.obs["{self.result_key}"]')
        test_x = adata.layers[self.layer_key] if self.layer_key else adata.X

        if adata.uns["_prediction_mode"] == "retrain":
            train_idx = adata.obs["_ref_subsample"]
            train_x = adata[train_idx].layers[self.layer_key] if self.layer_key else adata[train_idx].X
            train_y = adata.obs.loc[train_idx, self.labels_key].cat.codes.to_numpy()

            if settings.cuml:
                from cuml.svm import LinearSVC
                from sklearn.multiclass import OneVsRestClassifier

                # Enable probability support if requested
                self.classifier_dict["probability"] = self.return_probabilities
                from cuml.svm import LinearSVC as cumlLinearSVC  # ensure import aliasing
                clf_cuml = OneVsRestClassifier(cumlLinearSVC(**self.classifier_dict))
                train_x_dense = train_x.todense() if hasattr(train_x, "todense") else train_x
                clf_cuml.fit(train_x_dense, train_y)
                with open(
                    os.path.join(adata.uns["_save_path_trained_models"], "svm_classifier_cuml.joblib"), "wb"
                ) as f:
                    joblib.dump(clf_cuml, f)
                self.classifier_dict.pop("probability")
            if not settings.cuml or self.train_both:
                clf = CalibratedClassifierCV(svm.LinearSVC(**self.classifier_dict))
                clf.fit(train_x, train_y)
                with open(
                    os.path.join(adata.uns["_save_path_trained_models"], "svm_classifier.joblib"), "wb"
                ) as f:
                    joblib.dump(clf, f)

        # Set up result dataframe with appropriate columns
        columns = [self.result_key]
        if self.return_probabilities:
            columns.append(f"{self.result_key}_probabilities")

        result_df = pd.DataFrame(index=adata.obs_names, columns=columns, dtype=float)
        result_df[self.result_key] = result_df[self.result_key].astype("object")

        if settings.cuml:
            with open(
                os.path.join(adata.uns["_save_path_trained_models"], "svm_classifier_cuml.joblib"), "rb"
            ) as f:
                clf = joblib.load(f)
            shard_size = int(settings.shard_size)
            for i in range(0, adata.n_obs, shard_size):
                tmp_x = test_x[i : i + shard_size]
                names_x = adata.obs_names[i : i + shard_size]
                tmp_x_dense = tmp_x.todense() if hasattr(tmp_x, "todense") else tmp_x
                preds = clf.predict(tmp_x_dense).astype(int)
                result_df.loc[names_x, self.result_key] = adata.uns["label_categories"][preds]
                if self.return_probabilities:
                    probs = np.max(clf.predict_proba(tmp_x_dense), axis=1).astype(float)
                    result_df.loc[names_x, f"{self.result_key}_probabilities"] = probs
        else:
            with open(
                os.path.join(adata.uns["_save_path_trained_models"], "svm_classifier.joblib"), "rb"
            ) as f:
                clf = joblib.load(f)
            preds = clf.predict(test_x)
            result_df[self.result_key] = adata.uns["label_categories"][preds]
            if self.return_probabilities:
                result_df[f"{self.result_key}_probabilities"] = np.max(clf.predict_proba(test_x), axis=1)

        # Initialize missing columns in adata.obs if needed
        for col in columns:
            if col not in adata.obs.columns:
                adata.obs[col] = (
                    pd.Series(dtype="float64")
                    if "probabilities" in col
                    else adata.uns["unknown_celltype_label"]
                )

        adata.obs.loc[adata.obs["_predict_cells"] == "relabel", result_df.columns] = result_df



from typing import Optional
import numpy as np
import pandas as pd
import scanpy as sc

from scipy import sparse
from scipy.spatial import distance_matrix
import anndata
from .._optimal_transport import cot_sparse
from .._optimal_transport import cot_combine_sparse
from .._optimal_transport import uot
from .._optimal_transport import usot

def pairwise_scc(X1, X2):
    X1 = X1.argsort(axis=1).argsort(axis=1)
    X2 = X2.argsort(axis=1).argsort(axis=1)
    X1 = (X1-X1.mean(axis=1, keepdims=True))/X1.std(axis=1, keepdims=True)
    X2 = (X2-X2.mean(axis=1, keepdims=True))/X2.std(axis=1, keepdims=True)
    sccmat = np.empty([X1.shape[0], X2.shape[0]], float)
    for i in range(X1.shape[0]):
        for j in range(X2.shape[0]):
            c = np.dot( X1[i,:], X2[j,:]) / float(X1.shape[1])
            sccmat[i,j] = c
    return sccmat

def infer_spatial_information(
    adata_sc: anndata.AnnData,
    adata_sp: anndata.AnnData,
    cost_sc_sp: Optional[np.ndarray] = None,
    cost_sc: Optional[np.ndarray] = None,
    cost_sp: Optional[np.ndarray] = None,
    ot_alpha: float = 0.2,
    ot_rho: float = 0.05,
    ot_epsilon: float = 0.01,
    exp_pred_prone: float = 0.0,
    loc_pred_k: int = 1,
    return_gamma: bool = False
):
    """
    Infer spatial information.

    Given a pair of spatial data and scRNA-seq data, estimate spatial origin
    of scRNA-seq data and impute gene expression for spatial data [Cang2020]_.

    Parameters
    ----------
    adata_sc
        The data matrix for scRNA-seq data of shape ``n_obs(sc)`` × ``n_vars(sc)``.
        Rows corresponds to cells and columns to genes.
    adata_sp
        The data matrix for spatial data of shape ``n_obs(sp)`` × ``n_vars(sp)``.
        Rows corresponds to positions and columns to genes.
    cost_sc_sp
        The dissimilarity matrix between scRNA-seq data and spatial data of shape ``n_obs(sc)`` × ``n_obs(sp)``.
        If not given, 1 - Spearman's r on common genes is used.
    cost_sc
        The dissimilarity matrix within scRNA-seq data of shape ``n_obs(sc)`` × ``n_obs(sc)``.
        Only needed when structured optimal transport is used (ot_alpha > 0).
        If not given, the Euclidean distance in PCA space is used.
    cost_sp
        The distance matrix within spatial data of shape ``n_obs(sp)`` × ``n_obs(sp)``.
        Only needed when structured optimal transport is used (ot_alpha > 0).
        If not given, the spatial distance among spatial locations is used.
    ot_alpha
        Weight for the structured component in optimal transport in [0,1]. 
    ot_rho
        Marginal relaxtion term (>0). Traditional OT when ot_rho=inf.
    ot_epsilon
        Entropy regularization term (>0). A higher value will generate a denser mapping matrix.
    exp_pred_prone
        The percentage of cells with low weights to ignore when predicing gene expression for each spatial data. 
        A higher percentage will increase the sparseness of the predicted spatial data due to the sparseness in the scRNA-seq data.
    loc_pred_k
        Number of top spatial matches for predicting spatial origin of cells. 
    return_gamma
        Whether to return the optimal transport plan (gamma matrix)

    Returns
    -------
    adata_sc_pred : anndata.AnnData
        The scRNA-seq data with predicted spatial origins in ``.obsm['spatial']``.
    adata_sp_pred : anndata.AnnData
        The spatial data with imputed gene expression.
    gamma : np.ndarray
        The connectivity matrix between scRNA-seq data and spatial data which is used as weights to generate the predicted datasets adata_sc_pred and adata_sp_pred.


    References
    ----------

    .. [Cang2020] Cang, Z., & Nie, Q. (2020). Inferring spatial and signaling 
        relationships between cells from single cell transcriptomic data. 
        Nature communications, 11(1), 1-13.

    """
    
    # Cost matrices for optimal transport
    if cost_sc_sp is None:
        common_genes = list(set(adata_sc.var_names).intersection(set(adata_sp.var_names)))
        adata_sc_common = adata_sc[:, common_genes]
        adata_sp_common = adata_sp[:, common_genes]
        cor_sc_sp = pairwise_scc(np.array(adata_sc_common.X.toarray()), np.array(adata_sp_common.X.toarray()))
        cost_sc_sp = 0.5 * ( 1.0 - cor_sc_sp )
    if cost_sc is None and ot_alpha != 0.0:
        adata_sc_processed = adata_sc.copy()
        sc.pp.highly_variable_genes(adata_sc_processed, n_top_genes=1000)
        adata_sc_processed = adata_sc_processed[:, adata_sc_processed.var.highly_variable]
        sc.pp.scale(adata_sc_processed, max_value=10)
        sc.tl.pca(adata_sc_processed, svd_solver='arpack', n_comps=20)
        x_pca = adata_sc_processed.obsm["X_pca"]
        cost_sc = distance_matrix(x_pca, x_pca)
    if cost_sp is None and ot_alpha != 0.0:
        cost_sp = distance_matrix(adata_sp.obsm["spatial"], adata_sp.obsm["spatial"])
    
    if ot_alpha != 0:
        cost_sp = cost_sp / np.max(cost_sp)
        cost_sc = cost_sc / np.max(cost_sc)
    cost_sc_sp = cost_sc_sp / np.max(cost_sc_sp)

    # Optimal transport mapping for the two datasets
    w_mat = np.exp(-cost_sc_sp)
    w_sc = np.sum(w_mat, axis=1); w_sp = np.sum(w_mat, axis=0)
    w_sc = w_sc / np.sum(w_sc); w_sp = w_sp / np.sum(w_sp)

    if ot_alpha == 0.0 and np.isinf(ot_rho):
        import ot
        gamma = ot.sinkhorn(w_sc, w_sp, cost_sc_sp, ot_epsilon)
    elif ot_alpha == 0.0 and not np.isinf(ot_rho):
        gamma = uot(w_sc, w_sp, cost_sc_sp, ot_epsilon, rho = ot_rho)
    else:
        gamma = usot(w_sc, w_sp, cost_sc_sp, cost_sc, cost_sp, ot_alpha,
                            epsilon = ot_epsilon, rho = ot_rho)

    # Spatial data with predicted gene expression
    gamma_sp = gamma.copy()
    if exp_pred_prone > 0.0:
        tmp_quantile = np.quantile(gamma_sp, exp_pred_prone, axis=0)
        tmp_gamma_sp = gamma_sp - tmp_quantile
        gamma_sp[np.where(tmp_gamma_sp < 0)] = 0
    gamma_sp = gamma_sp / gamma_sp.sum(axis=0)
    if isinstance(adata_sc.X, np.ndarray):
        X_sp_pred = np.matmul(gamma_sp.T, np.array(adata_sc.X))
    elif isinstance(adata_sc.X, sparse.csr_matrix):
        X_sp_pred = sparse.csr_matrix( gamma_sp.T * adata_sc.X )
    adata_sp_pred = anndata.AnnData(X=X_sp_pred, var=pd.DataFrame(index=adata_sc.var_names))
    adata_sp_pred.obsm["spatial"] = adata_sp.obsm["spatial"]

    # Single-cell data with predicted spatial location
    gamma_sc = gamma.copy()
    gamma_sc_sorted = -np.sort(-gamma_sc, axis=1)
    k = loc_pred_k
    for i in range(gamma_sc.shape[0]):
        gamma_sc[i,np.where(gamma_sc[i,:] < gamma_sc_sorted[i,k])[0]] = 0
        gamma_sc[i,:] = gamma_sc[i,:] / np.sum(gamma_sc[i,:])
    pos_pred = np.matmul(gamma_sc, adata_sp.obsm["spatial"])
    adata_sc_pred = adata_sc.copy()
    adata_sc_pred.obsm["spatial"] = pos_pred

    if return_gamma:
        return adata_sc_pred, adata_sp_pred, gamma
    else:
        return adata_sc_pred, adata_sp_pred



import numpy as np
from scipy import sparse
from scipy.spatial import distance_matrix
from scipy.special import wrightomega
import matplotlib.pyplot as plt

import torch
import numpy as np
import scipy.sparse as sp
from tqdm import trange

def unot(a,
        b,
        C,
        eps_p,
        rho,
        eps_mu=None,
        eps_nu=None,
        sparse_mtx=False,
        solver="sinkhorn",
        nitermax=10000,
        stopthr=1e-8,
        verbose=False,
        momentum_dt=0.1,
        momentum_beta=0.0):
    """ The main function calling different algorithms.

    Parameters
    ----------
    a : (ns,) numpy.ndarray
        Source distribution. The summation should be less than or equal to 1.
    b : (nt,) numpy.ndarray
        Target distribution. The summation should be less than or equal to 1.
    C : (ns,nt) numpy.ndarray
        The cost matrix possibly with infinity entries.
    eps_p :  float
        The coefficient of entropy regularization for P.
    rho : float
        The coefficient of penalty for unmatched mass.
    eps_mu : float, defaults to eps_p
        The coefficient of entropy regularization for mu.
    eps_nu : float, defaults to eps_p
        The coefficient of entropy regularization for nu.
    sparse_mtx : boolean, defaults to False
        Whether using sparse matrix format. If True, C should be in coo_sparse format.
    solver : str, defaults to 'sinkhorn'
        The solver to use. Choose from 'sinkhorn' and 'momentum'.
    nitermax : int, defaults to 10000
        The maximum number of iterations.
    stopthr : float, defaults to 1e-8
        The relative error threshold for stopping.
    verbose : boolean, defaults to False
        Whether to print algorithm logs.
    momentum_dt : float, defaults to 1e-1
        Step size if momentum method is used.
    momentum_beta : float, defautls to 0
        The coefficient for the momentum term if momemtum method is used.
    """
    if eps_mu is None: eps_mu = eps_p
    if eps_nu is None: eps_nu = eps_p
    # Return a zero matrix if either a or b is all zero
    nzind_a = np.where(a > 0)[0]; nzind_b = np.where(b > 0)[0]
    if len(nzind_a) == 0 or len(nzind_b) == 0:
        if sparse_mtx:
            P = sparse.coo_matrix(([],([],[])), shape=(len(a), len(b)))
        else:
            P = np.zeros([len(a), len(b)], float)
        return P
    if solver == "sinkhorn" and max(abs(eps_p-eps_mu),abs(eps_p-eps_nu))>1e-8:
        print("To use Sinkhorn algorithm, set eps_p=eps_mu=eps_nu")
        exit()
    if solver == "sinkhorn" and not sparse_mtx:
        P = unot_sinkhorn_l1_dense(a,b,C,eps_p,rho, \
            nitermax=nitermax,stopthr=stopthr,verbose=verbose)
    elif solver == "sinkhorn" and sparse_mtx:
        P = unot_sinkhorn_l1_sparse(a,b,C,eps_p,rho, \
            nitermax=nitermax,stopthr=stopthr,verbose=verbose)
    elif solver == "momentum" and not sparse_mtx: 
        P = unot_momentum_l1_dense(a,b,C,eps_p,eps_mu,eps_nu,rho, \
            nitermax=nitermax,stopthr=stopthr,dt=momentum_dt, \
            beta=momentum_beta,precondition=True,verbose=verbose)
    elif solver == "momentum" and sparse_mtx:
        print("under construction")
        exit()
    return P

def unot_sinkhorn_l2_dense(a,b,C,eps,m,nitermax=10000,stopthr=1e-8,verbose=False):
    """ Solve the unnormalized optimal transport with l2 penalty in dense matrix format.

    Parameters
    ----------
    a : (ns,) numpy.ndarray
        Source distribution. The summation should be less than or equal to 1.
    b : (nt,) numpy.ndarray
        Target distribution. The summation should be less than or equal to 1.
    C : (ns,nt) numpy.ndarray
        The cost matrix possibly with infinity entries.
    eps : float
        The coefficient for entropy regularization.
    m : float
        The coefficient for penalizing unmatched mass.
    nitermax : int, optional
        The max number of iterations. Defaults to 10000.
    stopthr : float, optional
        The threshold for terminating the iteration. Defaults to 1e-8.

    Returns
    -------
    (ns,nt) numpy.ndarray
        The optimal transport matrix.
    """
    f = np.zeros_like(a)
    g = np.zeros_like(b)
    r = np.zeros_like(a)
    s = np.zeros_like(b)
    niter = 0
    err = 100
    while niter <= nitermax and err > stopthr:
        fprev = f
        gprev = g
        # Iteration
        f = eps * np.log(a) \
            - eps * np.log( np.sum( np.exp( ( f.reshape(-1,1) + g.reshape(1,-1) - C ) / eps ), axis=1 ) \
            + np.exp( ( r + f ) / eps ) ) + f
        g = eps * np.log(b) \
            - eps * np.log( np.sum( np.exp( ( f.reshape(-1,1) + g.reshape(1,-1) - C ) / eps ), axis=0 ) \
            + np.exp( ( s + g ) / eps ) ) + g
        r = - eps * wrightomega( f/eps - np.log( eps/m ) ).real
        s = - eps * wrightomega( g/eps - np.log( eps/m ) ).real
        # Check relative error
        if niter % 10 == 0:
            err_f = abs(f - fprev).max() / max(abs(f).max(), abs(fprev).max(), 1.)
            err_g = abs(g - gprev).max() / max(abs(g).max(), abs(gprev).max(), 1.)
            err = 0.5 * (err_f + err_g)
        niter = niter + 1
    if verbose:
        print('Number of iterations in unot:', niter)
    P = np.exp( ( f.reshape(-1,1) + g.reshape(1,-1) - C ) / eps )
    return P

def unot_sinkhorn_l1_dense(a,b,C,eps,m,nitermax=10000,stopthr=1e-8,verbose=False,output_fg=False):
    """ Solve the unnormalized optimal transport with l1 penalty in dense matrix format.

    Parameters
    ----------
    a : (ns,) numpy.ndarray
        Source distribution. The summation should be less than or equal to 1.
    b : (nt,) numpy.ndarray
        Target distribution. The summation should be less than or equal to 1.
    C : (ns,nt) numpy.ndarray
        The cost matrix possibly with infinity entries.
    eps : float
        The coefficient for entropy regularization.
    m : float
        The coefficient for penalizing unmatched mass.
    nitermax : int, optional
        The max number of iterations. Defaults to 10000.
    stopthr : float, optional
        The threshold for terminating the iteration. Defaults to 1e-8.

    Returns
    -------
    (ns,nt) numpy.ndarray
        The optimal transport matrix.
    """
    f = np.zeros_like(a)
    g = np.zeros_like(b)
    niter = 0
    err = 100
    while niter <= nitermax and err >  stopthr:
        fprev = f
        gprev = g
        # Iteration
        f = eps * np.log(a) \
            - eps * np.log( np.sum( np.exp( ( f.reshape(-1,1) + g.reshape(1,-1) - C ) / eps ), axis=1 ) \
            + np.exp( ( -m + f ) / eps ) ) + f
        g = eps * np.log(b) \
            - eps * np.log( np.sum( np.exp( ( f.reshape(-1,1) + g.reshape(1,-1) - C ) / eps ), axis=0 ) \
            + np.exp( ( -m + g ) / eps ) ) + g
        # Check relative error
        if niter % 10 == 0:
            err_f = abs(f - fprev).max() / max(abs(f).max(), abs(fprev).max(), 1.)
            err_g = abs(g - gprev).max() / max(abs(g).max(), abs(gprev).max(), 1.)
            err = 0.5 * (err_f + err_g)
        niter = niter + 1
    if verbose:
        print('Number of iterations in unot:', niter)
    P = np.exp( ( f.reshape(-1,1) + g.reshape(1,-1) - C ) / eps )
    if output_fg:
        return f,g
    else:
        return P

def unot_barycenter_sinkhorn_l1_dense(a,C,eps,m,w,nitermax=5):
    L = len(a)
    n = len(a[0])
    f,g = {},{}
    for k in range(L):
        f[k] = np.zeros([n],float)
        g[k] = np.zeros([n],float)

    for i in range(nitermax):
        for k in range(L):
            f[k] = eps * np.log(a[k]) \
                - eps * np.log( np.sum( np.exp( ( f[k].reshape(-1,1) + g[k].reshape(1,-1) - C ) / eps ), axis=1 ) #) + f[k] #\
                + np.exp( ( -m + f[k] ) / eps ) ) + f[k]
        Lnu = np.zeros([n],float)
        for k in range(L):
            Lnu = Lnu + w[k] * np.log( np.sum( np.exp( ( f[k].reshape(-1,1) + g[k].reshape(1,-1) - C ) / eps ), axis=0 ))
        for k in range(L):
            g[k] = eps * Lnu \
                - eps * np.log( np.sum( np.exp( ( f[k].reshape(-1,1) + g[k].reshape(1,-1) - C ) / eps ), axis=0 ) #) + g[k] #\
                + np.exp( ( -m + g[k] ) / eps ) ) + g[k]
    return(np.exp(Lnu))

def regular_barycenter(a,C,eps,w,nitermax=10000):
    L = len(a)
    n = len(a[0])
    f,g = {},{}
    for k in range(L):
        f[k] = np.zeros([n],float)
        g[k] = np.zeros([n],float)

    for i in range(nitermax):
        for k in range(L):
            f[k] = eps * np.log(a[k]) \
                - eps * np.log( np.sum( np.exp( ( f[k].reshape(-1,1) + g[k].reshape(1,-1) - C ) / eps ), axis=1 ) ) + f[k]
        Lnu = np.zeros([n],float)
        for k in range(L):
            Lnu = Lnu + w[k] * np.log( np.sum( np.exp( ( f[k].reshape(-1,1) + g[k].reshape(1,-1) - C ) / eps ), axis=0 ))
        for k in range(L):
            g[k] = eps * Lnu \
                - eps * np.log( np.sum( np.exp( ( f[k].reshape(-1,1) + g[k].reshape(1,-1) - C ) / eps ), axis=0 ) ) + g[k]
    return(np.exp(Lnu))

def unot_sinkhorn_l2_sparse(a,b,C,eps,m,nitermax=10000,stopthr=1e-8,verbose=False):
    """ Solve the unnormalized optimal transport with l2 penalty in sparse matrix format.

    Parameters
    ----------
    a : (ns,) numpy.ndarray
        Source distribution. The summation should be less than or equal to 1.
    b : (nt,) numpy.ndarray
        Target distribution. The summation should be less than or equal to 1.
    C : (ns,nt) scipy.sparse.coo_matrix
        The cost matrix in coo sparse format. The entries exceeds the cost cutoff are omitted. The naturally zero entries should be explicitely included.
    eps : float
        The coefficient for entropy regularization.
    m : float
        The coefficient for penalizing unmatched mass.
    nitermax : int, optional
        The max number of iterations. Defaults to 10000.
    stopthr : float, optional
        The threshold for terminating the iteration. Defaults to 1e-8.

    Returns
    -------
    (ns,nt) scipy.sparse.coo_matrix
        The optimal transport matrix. The locations of entries should agree with C and there might by explicit zero entries.
    """
    tmp_K = C.copy()
    f = np.zeros_like(a)
    g = np.zeros_like(b)
    r = np.zeros_like(a)
    s = np.zeros_like(b)
    niter = 0
    err = 100
    while niter <= nitermax and err > stopthr:
        fprev = f
        gprev = g
        # Iteration
        tmp_K.data = np.exp( ( -C.data + f[C.row] + g[C.col] ) / eps )
        f = eps * np.log(a) \
            - eps * np.log( np.sum( tmp_K, axis=1 ).A.reshape(-1) \
            + np.exp( ( r + f ) / eps ) ) + f
        tmp_K.data = np.exp( ( -C.data + f[C.row] + g[C.col] ) / eps )
        g = eps * np.log(b) \
            - eps * np.log( np.sum( tmp_K, axis=0 ).A.reshape(-1) \
            + np.exp( ( s + g ) / eps ) ) + g
        r = - eps * wrightomega( f/eps - np.log( eps/m ) ).real
        s = - eps * wrightomega( g/eps - np.log( eps/m ) ).real
        # Check relative error
        if niter % 10 == 0:
            err_f = abs(f - fprev).max() / max(abs(f).max(), abs(fprev).max(), 1.)
            err_g = abs(g - gprev).max() / max(abs(g).max(), abs(gprev).max(), 1.)
            err = 0.5 * (err_f + err_g)
        niter = niter + 1
    if verbose:
        print('Number of iterations in unot:', niter)
    tmp_K.data = np.exp( ( -C.data + f[C.row] + g[C.col] ) / eps )
    return tmp_K

def _unot_sinkhorn_l1_sparse(a,b,C,eps,m,nitermax=10000,stopthr=1e-8,verbose=True):
    """ Solve the unnormalized optimal transport with l1 penalty in sparse matrix format.

    Parameters
    ----------
    a : (ns,) numpy.ndarray
        Source distribution. The summation should be less than or equal to 1.
    b : (nt,) numpy.ndarray
        Target distribution. The summation should be less than or equal to 1.
    C : (ns,nt) scipy.sparse.coo_matrix
        The cost matrix in coo sparse format. The entries exceeds the cost cutoff are omitted. The naturally zero entries should be explicitely included.
    eps : float
        The coefficient for entropy regularization.
    m : float
        The coefficient for penalizing unmatched mass.
    nitermax : int, optional
        The max number of iterations. Defaults to 10000.
    stopthr : float, optional
        The threshold for terminating the iteration. Defaults to 1e-8.

    Returns
    -------
    (ns,nt) scipy.sparse.coo_matrix
        The optimal transport matrix. The locations of entries should agree with C and there might by explicit zero entries.
    """
    tmp_K = C.copy()
    f = np.zeros_like(a)
    g = np.zeros_like(b)
    r = np.zeros_like(a)
    s = np.zeros_like(b)
    niter = 0
    err = 100
    from tqdm import tqdm
    for i in tqdm(range(nitermax),desc='unot: '):
        fprev = f
        gprev = g
        # Iteration
        tmp_K.data = np.exp( ( -C.data + f[C.row] + g[C.col] ) / eps )
        f = eps * np.log(a) \
            - eps * np.log( np.sum( tmp_K, axis=1 ).A.reshape(-1) \
            + np.exp( ( -m + f ) / eps ) ) + f
        tmp_K.data = np.exp( ( -C.data + f[C.row] + g[C.col] ) / eps )
        g = eps * np.log(b) \
            - eps * np.log( np.sum( tmp_K, axis=0 ).A.reshape(-1) \
            + np.exp( ( -m + g ) / eps ) ) + g
        # Check relative error
        if niter % 10 == 0:
            err_f = abs(f - fprev).max() / max(abs(f).max(), abs(fprev).max(), 1.)
            err_g = abs(g - gprev).max() / max(abs(g).max(), abs(gprev).max(), 1.)
            err = 0.5 * (err_f + err_g)
        niter = niter + 1
        if err <= stopthr:
            break
    

    if verbose:
        print('Number of iterations in unot:', niter)
    tmp_K.data = np.exp( ( -C.data + f[C.row] + g[C.col] ) / eps )
    return tmp_K

def unot_sinkhorn_l1_sparse(a, b, C_coo, eps, m, nitermax=10000, stopthr=1e-8, verbose=True, device='cuda'):
    """
    使用 Torch 在 GPU 上进行加速的 unot_sinkhorn_l1_sparse 版本。

    参数：
      a, b      : 一维 numpy 数组，表示源、目标分布（要求所有元素严格正）。
      C_coo     : scipy.sparse.coo_matrix，包含行索引、列索引和数据（成本矩阵）。
      eps, m    : 算法参数。
      nitermax  : 最大迭代次数。
      stopthr   : 收敛阈值。
      verbose   : 是否打印信息。
      device    : 使用的设备，默认 'cuda'，若无 GPU 可设置为 'cpu'。

    返回：
      result: 返回一个稀疏矩阵（这里以 torch.sparse_coo_tensor 的形式返回，如果需要最终转换成 SciPy 矩阵，可再做转换）。
    """
    dtype = torch.float64  # 建议使用 double 以提高精度
    # 将输入分布转换为 torch 张量并搬到指定设备
    a_torch = torch.tensor(a, dtype=dtype, device=device)
    b_torch = torch.tensor(b, dtype=dtype, device=device)
    
    ns = a.shape[0]
    nt = b.shape[0]
    
    # 将 COO 矩阵的行、列索引和成本数据转换为 torch 张量
    # 注意：这里 row, col 的数据类型应为 torch.long
    row = torch.tensor(C_coo.row, dtype=torch.long, device=device)
    col = torch.tensor(C_coo.col, dtype=torch.long, device=device)
    C_data = torch.tensor(C_coo.data, dtype=dtype, device=device)
    
    # 初始化 f 和 g，在 GPU 上
    f = torch.zeros(ns, dtype=dtype, device=device)
    g = torch.zeros(nt, dtype=dtype, device=device)
    
    eps0 = 1e-300  # 避免 log0
    err = torch.tensor(100., dtype=dtype, device=device)
    niter = 0
    
    # 主循环，采用 trange 展示进度条
    for i in trange(nitermax, desc="unot_torch", leave=False):
        f_prev = f.clone()
        g_prev = g.clone()
        
        # tmp = (-C_data + f[row] + g[col]) / eps
        tmp = (-C_data + f[row] + g[col]) / eps
        tmp_K = torch.exp(tmp)
        
        # 使用 torch.bincount 进行行、列聚合，注意要指定 minlength 保证输出维度
        # 对于行聚合:
        sum_row = torch.bincount(row, weights=tmp_K, minlength=ns)
        # 对于列聚合:
        sum_col = torch.bincount(col, weights=tmp_K, minlength=nt)
        
        # 更新 f 和 g
        f = eps * torch.log(a_torch + eps0) - eps * torch.log(sum_row + torch.exp((-m + f) / eps)) + f
        g = eps * torch.log(b_torch + eps0) - eps * torch.log(sum_col + torch.exp((-m + g) / eps)) + g
        
        # 每 10 次迭代计算连接误差
        if niter % 10 == 0:
            max_val = torch.max(torch.abs(f).max(), torch.abs(f_prev).max())
            max_val = torch.max(max_val, torch.tensor(1., dtype=dtype, device=device))
            err_f = torch.abs(f - f_prev).max() / max_val

            max_val_g = torch.max(torch.abs(g).max(), torch.abs(g_prev).max())
            max_val_g = torch.max(max_val_g, torch.tensor(1., dtype=dtype, device=device))
            err_g = torch.abs(g - g_prev).max() / max_val_g
            
            #err_f = torch.abs(f - f_prev).max() / torch.max(torch.abs(f), torch.abs(f_prev), torch.tensor(1., dtype=dtype, device=device))
            #err_g = torch.abs(g - g_prev).max() / torch.max(torch.abs(g), torch.abs(g_prev), torch.tensor(1., dtype=dtype, device=device))
            err = 0.5 * (err_f + err_g)
            if err.item() < stopthr:
                if verbose:
                    print("达到停止阈值，err =", err.item())
                break
        
        niter += 1
    
    if verbose:
        print("GPU 迭代次数：", niter, "最终误差：", err.item())
    
    # 最终计算 tmp_K 数据
    tmp_final = (-C_data + f[row] + g[col]) / eps
    tmp_K_final = torch.exp(tmp_final)
    
    # 构造 torch 的稀疏矩阵：indices 为 2 x N 的张量，values 为对应的数据
    indices = torch.stack([row, col], dim=0)  # shape: [2, nnz]
    result = torch.sparse_coo_tensor(indices, tmp_K_final, size=(ns, nt))
    
    # 如果需要将结果转为 CPU 上 SciPy 的稀疏矩阵，可以先转为 dense numpy 数组，但当矩阵大时要注意内存占用
    # 如下代码将 result 转至 CPU，并构造成 COO 矩阵：
    result_cpu = result.coalesce()
    values_cpu = result_cpu.values().cpu().numpy()
    indices_cpu = result_cpu.indices().cpu().numpy()
    scipy_result = sp.coo_matrix((values_cpu, (indices_cpu[0], indices_cpu[1])), shape=(ns, nt))
    return scipy_result

    #return result

def unot_nesterov_l2_dense(a,b,C,eps1,eps2,m,nitermax=10000,stopthr=1e-8):
    dt = 0.01
    f = np.zeros_like(a)
    g = np.zeros_like(b)
    f_old = np.array(f)
    g_old = np.array(g)
    F_old = np.array(f)
    G_old = np.array(g)
    niter = 0
    err = 100
    def Qf(ff,gg,ee1,ee2,mm,aa,bb,CC):
        out = np.exp(ff/ee1) * np.sum(np.exp((gg.reshape(1,-1)-CC)/ee1), axis=1) \
            + ee2 * wrightomega( ff*mm/ee2 - np.log( ee2/mm ) ).real / mm - aa
        return out
    def Qg(ff,gg,ee1,ee2,mm,aa,bb,CC):
        out = np.exp(gg/ee1) * np.sum(np.exp((ff.reshape(-1,1)-CC)/ee1), axis=0) \
            + ee2 * wrightomega( gg*mm/ee2 - np.log( ee2/mm ) ).real / mm - bb
        return out
    while niter <= nitermax and err > stopthr:
        f = F_old - dt * Qf(F_old, G_old, eps1, eps2, m, a, b, C)
        F = f + float(niter)/(niter+3.0) * (f-f_old)
        g = G_old - dt * Qg(F_old, G_old, eps1, eps2, m, a, b, C)
        G = g + float(niter)/(niter+3.0) * (g-g_old)
        if niter % 10 == 0:
            err_f = abs(f - f_old).max() / max(abs(f).max(), abs(f_old).max(), 1.)
            err_g = abs(g - g_old).max() / max(abs(g).max(), abs(g_old).max(), 1.)
            err = 0.5 * (err_f + err_g)
        f_old[:] = f[:]; F_old[:] = F[:]
        g_old[:] = g[:]; G_old[:] = G[:]
        niter += 1
    P = np.exp((f.reshape(-1,1)+g.reshape(1,-1)-C)/eps1)
    return P

def unot_momentum_l2_dense(a,b,C,eps_p,eps_mu,eps_nu,m,nitermax=1e4,stopthr=1e-8,dt=0.01,beta=0.8):
    f = np.zeros_like(a)
    g = np.zeros_like(b)
    f_old = np.array(f)
    g_old = np.array(g)
    F_old = np.array(f)
    G_old = np.array(g)
    niter = 0
    err = 100
    def Qf(ff,gg,ee_p,ee_mu,ee_nu,mm,aa,bb,CC):
        out = np.exp(ff/ee_p) * np.sum(np.exp((gg.reshape(1,-1)-CC)/ee_p), axis=1) \
            + ee_mu * wrightomega( ff/ee_mu - np.log( ee_mu/mm ) ).real / mm - aa
        return out
    def Qg(ff,gg,ee_p,ee_mu,ee_nu,mm,aa,bb,CC):
        out = np.exp(gg/ee_p) * np.sum(np.exp((ff.reshape(-1,1)-CC)/ee_p), axis=0) \
            + ee_nu * wrightomega( gg/ee_nu - np.log( ee_nu/mm ) ).real / mm - bb
        return out
    while niter <= nitermax and err > stopthr:
        F = beta * F_old + Qf(f_old, g_old, eps_p, eps_mu, eps_nu, m, a, b, C)
        f = f_old - dt * F
        G = beta * G_old + Qg(f_old, g_old, eps_p, eps_mu, eps_nu, m, a, b, C)
        g = g_old - dt * G
        f_old[:] = f[:]; F_old[:] = F[:]
        g_old[:] = g[:]; G_old[:] = G[:]
        niter += 1
    P = np.exp((f.reshape(-1,1)+g.reshape(1,-1)-C)/eps_p)
    return P

def unot_momentum_l1_dense(a,b,C,eps_p,eps_mu,eps_nu,m,nitermax=1e4,stopthr=1e-8,dt=0.01,beta=0.8,precondition=False,verbose=False):
    f = np.zeros_like(a)
    g = np.zeros_like(b)
    if precondition:
        f,g = unot_sinkhorn_l1_dense(a,b,C,eps_p,m,output_fg=True)
    f_old = np.array(f)
    g_old = np.array(g)
    F_old = np.array(f)
    G_old = np.array(g)
    niter = 0
    err = 100
    def Qf(ff,gg,ee_p,ee_mu,ee_nu,mm,aa,bb,CC):
        out = np.exp(ff/ee_p) * np.sum(np.exp((gg.reshape(1,-1)-CC)/ee_p), axis=1) \
            + np.exp((ff-mm)/ee_mu) - aa
        return out
    def Qg(ff,gg,ee_p,ee_mu,ee_nu,mm,aa,bb,CC):
        out = np.exp(gg/ee_p) * np.sum(np.exp((ff.reshape(-1,1)-CC)/ee_p), axis=0) \
            + np.exp((gg-mm)/ee_nu) - bb
        return out
    while niter <= nitermax and err > stopthr:
        F = beta * F_old + Qf(f_old, g_old, eps_p, eps_mu, eps_nu, m, a, b, C)
        f = f_old - dt * F
        G = beta * G_old + Qg(f_old, g_old, eps_p, eps_mu, eps_nu, m, a, b, C)
        g = g_old - dt * G
        if niter % 10 == 0:
            err_f = abs(f - f_old).max() / max(abs(f).max(), abs(f_old).max(), 1.)
            err_g = abs(g - g_old).max() / max(abs(g).max(), abs(g_old).max(), 1.)
            err = 0.5 * (err_f + err_g)
        f_old[:] = f[:]; F_old[:] = F[:]
        g_old[:] = g[:]; G_old[:] = G[:]
        niter += 1
    P = np.exp((f.reshape(-1,1)+g.reshape(1,-1)-C)/eps_p)
    if verbose:
        print(niter)
    return P

def unot_momentum_l1_2end_dense(a,b,C,eps_p,eps_mu,eps_nu,m,nitermax=1e4,stopthr=1e-8,dt=0.01,beta=0.8):
    f = np.zeros_like(a)
    g = np.zeros_like(b)
    f_old = np.array(f)
    g_old = np.array(g)
    F_old = np.array(f)
    G_old = np.array(g)
    niter = 0
    err = 100
    def Qf(ff,gg,ee_p,ee_mu,ee_nu,mm,aa,bb,CC):
        out = np.exp(ff/ee_p) * np.sum(np.exp((gg.reshape(1,-1)-CC)/ee_p), axis=1) \
            + aa / (np.exp(-(ff-mm)/ee_mu)+1) - aa
        return out
    def Qg(ff,gg,ee_p,ee_mu,ee_nu,mm,aa,bb,CC):
        out = np.exp(gg/ee_p) * np.sum(np.exp((ff.reshape(-1,1)-CC)/ee_p), axis=0) \
            + bb / (np.exp(-(gg-mm)/ee_nu)+1) - bb
        return out
    while niter <= nitermax and err > stopthr:
        F = beta * F_old + Qf(f_old, g_old, eps_p, eps_mu, eps_nu, m, a, b, C)
        f = f_old - dt * F
        G = beta * G_old + Qg(f_old, g_old, eps_p, eps_mu, eps_nu, m, a, b, C)
        g = g_old - dt * G
        f_old[:] = f[:]; F_old[:] = F[:]
        g_old[:] = g[:]; G_old[:] = G[:]
        niter += 1
    P = np.exp((f.reshape(-1,1)+g.reshape(1,-1)-C)/eps_p)
    print(niter)
    return P


import matplotlib.pyplot as plt
from tqdm import tqdm
import numpy as np
import pandas as pd
import scanpy as sc

def cell_type_smoothing(adata,use_rep='X_pca', N_neighbors=15,n_pcs=10,
                         weighted_smoothing=True, threshold=0.5,
                         col_name = 'cell_type_from_scMulan'):
    # 重新算一个nn，避免取的N_neighbors比画umap的时候选的参数大
    sc.pp.neighbors(adata,use_rep=use_rep, n_pcs=n_pcs, 
                    n_neighbors=N_neighbors, key_added="Smoothing")
    
    # smoothing_score是近邻平滑之后给的分数
    # cell_type_from_mulan_smoothing是平滑后的结果

    if type(threshold) == int:
        threshold = threshold/(N_neighbors-1)

    smoothing_score_list = []
    smoothing_celltype_list = []

    for idx in tqdm(range(adata.shape[0])):
        distances = np.array(adata.obsp['Smoothing_distances'][idx,:].todense()).reshape(-1)
        pre_labels = adata.obs[col_name][distances>0]

        if weighted_smoothing:
            weights = (1/distances[distances>0])/sum(1/distances[distances>0])
            max_score = 0
            result_celltype = ""
            for ct in pre_labels.unique():
                if weights[pre_labels==ct].sum()>max_score:
                    max_score = weights[pre_labels==ct].sum()
                    result_celltype = ct
            smoothing_celltype_list.append(result_celltype)
            smoothing_score_list.append(max_score)

        else:
            result_counter = pre_labels.value_counts()[pre_labels.value_counts()>0]
            smoothing_celltype_list.append(result_counter.index[result_counter.argmax()])
            smoothing_score_list.append(result_counter.max()/(N_neighbors-1))

    adata.obs['cell_type_from_mulan_smoothing'] = smoothing_celltype_list
    adata.obs['smoothing_score'] = smoothing_score_list
    adata.obs.cell_type_from_mulan_smoothing[adata.obs.smoothing_score<threshold] = "Unclassified"
    adata.obs[col_name]=adata.obs[col_name].astype('category')
    if not (col_name + '_colors' in adata.uns):
        if len(adata.obs[col_name].cat.categories)<102:
            adata.uns[col_name + '_colors'] = sc.pl.palettes.default_102[:len(adata.obs[col_name].cat.categories)]
        else:
            adata.uns[col_name + '_colors'] = ['grey' for _ in range(len(adata.obs[col_name].cat.categories))]
        #sc.pl.umap(adata,color=[col_name],show=False)
        #plt.close()
    
    adata.uns['cell_type_from_mulan_smoothing_colors'] = ["#666666" for _ in adata.obs['cell_type_from_mulan_smoothing'].unique()]
    adata.obs[col_name] = pd.Categorical(adata.obs[col_name])
    adata.obs.cell_type_from_mulan_smoothing = pd.Categorical(adata.obs.cell_type_from_mulan_smoothing)
    for ct in adata.obs['cell_type_from_mulan_smoothing'].unique():
        if ct != 'Unclassified':
            ct_index = np.where(adata.obs[col_name].cat.categories==ct)[0][0]
            r_index = np.where(adata.obs["cell_type_from_mulan_smoothing"].cat.categories==ct)[0][0]
            adata.uns['cell_type_from_mulan_smoothing_colors'][r_index] = adata.uns[col_name + '_colors'][ct_index]


def visualize_selected_cell_types(adata, selected_cell_types, smoothing=False, col_name = 'cell_type_from_scMulan', **kwargs):
    if smoothing:
        selected_column = "cell_type_from_mulan_smoothing"
        if not (selected_column in adata.obs.columns):
            "You set 'smoothing=True', please run cell_type_smoothing before visualization."
            return -1
    else:
        selected_column = col_name

    adata.obs["selected_celltype"] = adata.obs[selected_column].astype(str)
    adata.obs.loc[[not (x in selected_cell_types) for x in adata.obs[selected_column]],"selected_celltype"] = "others"
    adata.obs.selected_celltype = pd.Categorical(adata.obs["selected_celltype"])
    adata.uns['selected_celltype_colors'] = ["#666666" for _ in adata.obs['selected_celltype'].unique()]
    # adata.obs.cell_type = pd.Categorical(adata.obs[selected_column])
    
    for ct in adata.obs['selected_celltype'].unique():
        if ct != 'others':
            ct_index = np.where(adata.obs[selected_column].cat.categories==ct)[0][0]
            r_index = np.where(adata.obs["selected_celltype"].cat.categories==ct)[0][0]
            adata.uns['selected_celltype_colors'][r_index] = adata.uns['{}_colors'.format(selected_column)][ct_index]
    print(adata.obs.selected_celltype.value_counts())
    sc.pl.umap(adata,color=["selected_celltype"], **kwargs)


import torch
from torch.nn import Embedding, Linear, LeakyReLU, PReLU, Parameter

class NTD(torch.nn.Module):
    def __init__(self, n_x, n_y, n_g, rank, random_state=1234567):
        super().__init__()
        torch.manual_seed(random_state)
        
        # Define embedding layer along x, y, g modes
        self.embedding_x = Embedding(n_x, rank)
        self.embedding_y = Embedding(n_y, rank)
        self.embedding_g = Embedding(n_g, rank)
        # Define nonlinear mapping layer along x, y, g modes
        self.lin_x_1 = Linear(rank, rank)
        self.lin_y_1 = Linear(rank, rank)
        self.lin_g_1 = Linear(rank, rank)
        self.prelu = PReLU(init=0.9)

    def forward(self, x_index, y_index, g_index):
        
        # Linear factors
        x = self.embedding_x(x_index)
        y = self.embedding_y(y_index)
        g = self.embedding_g(g_index)
        
        # Nonlinear factors
        x = self.lin_x_1(x)
        x = self.prelu(x)
        y = self.lin_y_1(y)
        y = self.prelu(y)
        g = self.lin_g_1(g)
        g = self.prelu(g)
        
        # Nonlinear aggregation 
        o = torch.einsum('im,jm,km->ijk', g, x, y)
        o = o.relu_()
        
        return x, y, g, o


from __future__ import annotations

import logging

import numpy as np
import scanpy as sc
from pynndescent import PyNNDescentTransformer
from sklearn.neighbors import KNeighborsClassifier
from sklearn.pipeline import make_pipeline

from popv import settings
from popv.algorithms._base_algorithm import BaseAlgorithm


class KNN_SCANORAMA(BaseAlgorithm):
    """
    Compute KNN classifier after Scanorama integration.

    Parameters
    ----------
    batch_key : str, optional
        Key in adata.obs containing batch information. Default is "_batch_annotation".
    labels_key : str, optional
        Key in adata.obs containing cell-type labels. Default is "_labels_annotation".
    result_key : str, optional
        Key in adata.obs where predictions are stored.
        Default is "popv_knn_scanorama_prediction".
    embedding_key : str, optional
        Key in adata.obsm with the integrated embedding.
        Default is "X_pca_scanorama_popv".
    umap_key : str, optional
        Key in adata.obsm with the UMAP embedding.
        Default is "X_umap_scanorama_popv".
    method_kwargs : dict, optional
        Additional parameters for Scanorama integration.
    classifier_kwargs : dict, optional
        KNeighborsClassifier parameters.
        Default is {"weights": "uniform", "n_neighbors": 15}.
    embedding_kwargs : dict, optional
        Additional parameters for UMAP embedding. Default is {"min_dist": 0.1}.
    """

    def __init__(
        self,
        batch_key: str | None = "_batch_annotation",
        labels_key: str | None = "_labels_annotation",
        result_key: str | None = "popv_knn_scanorama_prediction",
        embedding_key: str | None = "X_pca_scanorama_popv",
        umap_key: str | None = "X_umap_scanorama_popv",
        method_kwargs: dict | None = None,
        classifier_kwargs: dict | None = None,
        embedding_kwargs: dict | None = None,
    ) -> None:
        super().__init__(
            batch_key=batch_key,
            labels_key=labels_key,
            result_key=result_key,
            embedding_key=embedding_key,
            umap_key=umap_key,
        )

        self.method_kwargs = method_kwargs or {}
        self.classifier_kwargs = {"weights": "uniform", "n_neighbors": 15}
        self.classifier_kwargs.update(classifier_kwargs or {})
        self.embedding_kwargs = {"min_dist": 0.1}
        self.embedding_kwargs.update(embedding_kwargs or {})

    def compute_integration(self, adata):
        """
        Integrate data using Scanorama.

        Parameters
        ----------
        adata : AnnData
            AnnData object. Integrated embedding is stored in adata.obsm[self.embedding_key].
        """
        logging.info("Integrating data with scanorama")
        # Sort and copy adata for integration
        sorted_idx = adata.obs.sort_values(self.batch_key).index
        tmp = adata[sorted_idx].copy()
        sc.external.pp.scanorama_integrate(
            tmp,
            key=self.batch_key,
            adjusted_basis=self.embedding_key,
            **self.method_kwargs,
        )
        adata.obsm[self.embedding_key] = tmp[adata.obs_names].obsm[self.embedding_key]

    def predict(self, adata):
        """
        Compute KNN classifier on Scanorama integrated data.

        Parameters
        ----------
        adata : AnnData
            AnnData object. Predictions are stored in adata.obs[self.result_key].
        """
        logging.info(f'Saving knn predictions to adata.obs["{self.result_key}"]')

        ref_idx = adata.obs["_labelled_train_indices"]
        train_X = adata[ref_idx].obsm[self.embedding_key]
        train_Y = adata.obs.loc[ref_idx, self.labels_key].cat.codes.to_numpy()

        knn = make_pipeline(
            PyNNDescentTransformer(
                n_neighbors=self.classifier_kwargs["n_neighbors"],
                n_jobs=settings.n_jobs,
            ),
            KNeighborsClassifier(
                metric="precomputed", weights=self.classifier_kwargs["weights"]
            ),
        )

        knn.fit(train_X, train_Y)
        knn_pred = knn.predict(adata.obsm[self.embedding_key])
        adata.obs[self.result_key] = adata.uns["label_categories"][knn_pred]

        if getattr(self, "return_probabilities", False):
            probs = knn.predict_proba(adata.obsm[self.embedding_key])
            adata.obs[f"{self.result_key}_probabilities"] = np.max(probs, axis=1)

    def compute_umap(self, adata):
        """
        Compute UMAP embedding of integrated data.

        Parameters
        ----------
        adata : AnnData
            AnnData object. UMAP embedding is stored in adata.obsm[self.umap_key].
        """
        if getattr(self, "compute_umap_embedding", False):
            logging.info(f'Saving UMAP embedding to adata.obsm["{self.umap_key}"]')
            transformer = "rapids" if settings.cuml else None
            sc.pp.neighbors(
                adata, use_rep=self.embedding_key, transformer=transformer
            )
            method = "rapids" if settings.cuml else "umap"
            umap_result = sc.tl.umap(adata, copy=True, method=method, **self.embedding_kwargs)
            adata.obsm[self.umap_key] = umap_result.obsm["X_umap"]



"""Utilities for translating bulk data to single-cell context."""

from ._bulk2single import Bulk2Single
from ._single2spatial import Single2Spatial
from ._bulktrajblend import BulkTrajBlend
from ._utils import bulk2single_plot_cellprop,bulk2single_plot_correlation

__all__ = [
    'Bulk2Single',
    'Single2Spatial',
    'BulkTrajBlend',
    'bulk2single_plot_cellprop',
    'bulk2single_plot_correlation',
]



import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils
import torch.distributions
import numpy as np

def positional_encoding(coords, enc_dim, sigma):
    freqs = (
        2 * np.pi * sigma ** (torch.arange(enc_dim//2, dtype=torch.float, device=coords.device) / enc_dim)
    )
    freqs = torch.reshape(freqs, (1,1, torch.numel(freqs)))
    coords = coords.unsqueeze(-1)
    
    freqs = coords * freqs #N x 2 x enc_dim/2
    s = torch.sin(freqs)
    c = torch.cos(freqs)

    x = torch.cat((s,c), axis=-1) #N x 2 x enc_dim
    x = torch.reshape(x, (x.shape[0], -1))

    return x